#!/usr/bin/env python
"""Guiguts - an application to support creation of ebooks for PG"""


import argparse
import logging
import importlib.resources
from importlib.metadata import version
import os.path
from tkinter import messagebox
from typing import Optional
import unicodedata
import webbrowser

from guiguts.data import themes
from guiguts.file import File, the_file, NUM_RECENT_FILES
from guiguts.footnotes import footnote_check, FootnoteIndexStyle
from guiguts.html_convert import HTMLGeneratorDialog, HTMLMarkupTypes
from guiguts.html_tools import HTMLImageDialog, html_validator_check, html_link_check
from guiguts.illo_sn_fixup import illosn_check
from guiguts.maintext import maintext
from guiguts.mainwindow import (
    MainWindow,
    Menu,
    menubar,
    StatusBar,
    statusbar,
    ErrorHandler,
    process_accel,
    mainimage,
    AutoImageState,
)
from guiguts.misc_dialogs import (
    PreferencesDialog,
    ComposeSequenceDialog,
    ComposeHelpDialog,
    UnicodeBlockDialog,
    UnicodeSearchDialog,
)
from guiguts.misc_tools import (
    basic_fixup_check,
    page_separator_fixup,
    PageSepAutoType,
    unmatched_dp_markup,
    unmatched_html_markup,
    unmatched_brackets,
    unmatched_curly_quotes,
    unmatched_block_markup,
    FractionConvertType,
    fraction_convert,
    unicode_normalize,
    proofer_comment_check,
    asterisk_check,
    TextMarkupConvertorDialog,
    stealth_scannos,
    DEFAULT_SCANNOS_DIR,
    DEFAULT_REGEX_SCANNOS,
    DEFAULT_STEALTH_SCANNOS,
    DEFAULT_MISSPELLED_SCANNOS,
    convert_to_curly_quotes,
    check_curly_quotes,
)
from guiguts.page_details import PageDetailsDialog
from guiguts.preferences import preferences, PrefKey
from guiguts.root import root
from guiguts.search import show_search_dialog, find_next
from guiguts.spell import spell_check
from guiguts.tools.bookloupe import bookloupe_check
from guiguts.tools.jeebies import jeebies_check, JeebiesParanoiaLevel
from guiguts.tools.levenshtein import levenshtein_check, LevenshteinEditDistance
from guiguts.tools.pptxt import pptxt
from guiguts.utilities import is_mac, folder_dir_str
from guiguts.widgets import themed_style, theme_name_internal_from_user
from guiguts.word_frequency import word_frequency, WFDisplayType, WFSortType

logger = logging.getLogger(__package__)

MESSAGE_FORMAT = "%(asctime)s: %(levelname)s - %(message)s"
DEBUG_FORMAT = "%(asctime)s: %(levelname)s - %(filename)s:%(lineno)d - %(message)s"
THEMES_DIR = importlib.resources.files(themes)


class Guiguts:
    """Top level Guiguts application."""

    def __init__(self) -> None:
        """Initialize Guiguts class.

        Creates windows and sets default preferences."""

        self.parse_args()

        self.logging_init()
        logger.info("Guiguts started")

        self.initialize_preferences()

        self.file = File(self.filename_changed, self.languages_changed)
        the_file(self.file)

        self.mainwindow = MainWindow()
        self.update_title()

        theme_path = THEMES_DIR.joinpath("awthemes-10.4.0")
        root().tk.call("lappend", "auto_path", theme_path)
        root().tk.call("package", "require", "awdark")
        root().tk.call("package", "require", "awlight")

        self.file_menu: Optional[Menu] = (
            None  # File menu is saved to allow deletion & re-creation
        )
        self.init_menus()

        self.init_statusbar(statusbar())

        self.file.languages = preferences.get(PrefKey.DEFAULT_LANGUAGES)

        maintext().focus_set()
        maintext().add_modified_callback(self.update_title)
        maintext().add_config_callback(self.update_title)

        # Known tkinter issue - must call this before any dialogs can get created,
        # or focus will not return to maintext on Windows
        root().update_idletasks()

        # After menus, etc., have been created, set the zoomed/fullscreen state
        root().set_zoom_fullscreen()

        self.logging_add_gui()
        logger.info("GUI initialized")

        preferences.run_callbacks()

        self.load_file_if_given()

        def check_save_and_destroy() -> None:
            if self.file.check_save():
                root().destroy()

        root().protocol("WM_DELETE_WINDOW", check_save_and_destroy)

    def parse_args(self) -> None:
        """Parse command line args"""
        parser = argparse.ArgumentParser(
            prog="guiguts", description="Guiguts is an ebook creation tool"
        )
        parser.add_argument(
            "filename", nargs="?", help="Optional name of file to be loaded"
        )
        parser.add_argument(
            "-r",
            "--recent",
            type=int,
            choices=range(1, NUM_RECENT_FILES + 1),
            help="Number of 'Recent File' to be loaded: 1 is most recent",
        )
        parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="Run in debug mode",
        )
        parser.add_argument(
            "--nohome",
            action="store_true",
            help="Do not load the Preferences file",
        )
        self.args = parser.parse_args()

    def load_file_if_given(self) -> None:
        """If filename, or recent number, given on command line
        load the relevant file."""

        if self.args.filename:
            self.file.load_file(self.args.filename)
        elif self.args.recent:
            index = self.args.recent - 1
            try:
                self.file.load_file(preferences.get(PrefKey.RECENT_FILES)[index])
            except IndexError:
                pass  # Not enough recent files to load the requested one

    def auto_image_callback(self, value: bool) -> None:
        """Callback when auto_image preference is changed."""
        statusbar().set("see img", "Auto Img" if value else "See Img")
        if value:
            mainimage().auto_image_state(AutoImageState.NORMAL)
            self.image_dir_check()
            self.auto_image_check()

    def auto_image_check(self) -> None:
        """Function called repeatedly to check whether an image needs loading."""
        if preferences.get(PrefKey.AUTO_IMAGE):
            # Image viewer can temporarily pause auto image viewing,
            # but still need to schedule another call to this method.
            auto_image_state = mainimage().auto_image_state()
            if auto_image_state != AutoImageState.PAUSED:
                self.mainwindow.load_image(self.file.get_current_image_path())
                if auto_image_state == AutoImageState.RESTARTING:
                    mainimage().alert_user()
                    mainimage().auto_image_state(AutoImageState.NORMAL)

            root().after(200, self.auto_image_check)

    def show_image(self) -> None:
        """Show the image corresponding to current location."""
        self.image_dir_check()
        self.mainwindow.load_image(self.file.get_current_image_path())

    def hide_image(self) -> None:
        """Hide the image."""
        preferences.set(PrefKey.AUTO_IMAGE, False)
        self.mainwindow.hide_image()

    def image_dir_check(self) -> None:
        """Check if image dir is set up correctly."""
        if self.file.filename and not (
            self.file.image_dir and os.path.exists(self.file.image_dir)
        ):
            self.file.choose_image_dir()

    def run(self) -> None:
        """Run the app."""
        root().mainloop()

    def filename_changed(self) -> None:
        """Handle side effects needed when filename changes."""
        self.init_file_menu()  # Recreate file menu to reflect recent files
        self.update_title()
        if preferences.get(PrefKey.AUTO_IMAGE):
            self.image_dir_check()
        maintext().after_idle(maintext().focus_set)

    def languages_changed(self) -> None:
        """Handle side effects needed when languages change."""
        statusbar().set("languages label", "Lang: " + (self.file.languages or ""))

    def update_title(self) -> None:
        """Update the window title to reflect current status."""
        modtitle = " - edited" if maintext().is_modified() else ""
        filetitle = self.file.filename
        # Restrict filepath to rightmost W characters where W is maintext's visible width
        len_title = len(filetitle) + len(modtitle) + len("...")
        if len_title > 0:
            max_width = int(maintext().winfo_width() / maintext().font.measure("0"))
            if len_title > max_width:
                filetitle = "..." + filetitle[len_title - max_width :]
            if filetitle:
                filetitle = " - " + filetitle
        root().title(f"Guiguts {version('guiguts')}" + modtitle + filetitle)
        if is_mac():
            root().wm_attributes("-modified", maintext().is_modified())

    def quit_program(self) -> None:
        """Exit the program."""
        if self.file.check_save():
            root().quit()

    def help_about(self) -> None:
        """Display a 'Help About' dialog."""
        help_message = """Guiguts - an application to support creation of ebooks for PG

Copyright Contributors to the Guiguts-py project.

This program is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public
License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 51 Franklin Street,
Fifth Floor, Boston, MA 02110-1301 USA."""

        messagebox.showinfo(title="About Guiguts", message=help_message)

    def show_page_details_dialog(self) -> None:
        """Show the page details display/edit dialog."""
        PageDetailsDialog.show_dialog(page_details=self.file.page_details)

    def open_document(self, args: list[str]) -> None:
        """Handle drag/drop on Macs.

        Accepts a list of filenames, but only loads the first.
        """
        self.file.load_file(args[0])

    def open_file(self, filename: str = "") -> None:
        """Open new file, close old image if open.

        Args:
            filename: Optional filename - prompt user if none given.
        """
        if self.file.open_file(filename):
            self.mainwindow.clear_image()

    def close_command(self) -> None:
        """Close top-level dialog with focus, or close the file if it's
        the main window."""
        focus_widget = maintext().focus_get()
        if focus_widget is None:
            return
        focus_tl = focus_widget.winfo_toplevel()
        if focus_tl == maintext().winfo_toplevel():
            self.close_file()
        else:
            focus_tl.destroy()

    def close_file(self) -> None:
        """Close currently loaded file and associated image."""
        self.file.close_file()
        self.mainwindow.clear_image()

    def load_current_image(self) -> None:
        """Load image corresponding to current cursor position"""
        self.mainwindow.load_image(self.file.get_current_image_path())

    def show_help_manual(self) -> None:
        """Display the manual."""
        webbrowser.open("https://www.pgdp.net/wiki/PPTools/Guiguts/Guiguts_2_Manual")

    def show_help_regex(self) -> None:
        """Display the regex cheat sheet."""
        webbrowser.open(
            "https://www.pgdp.net/wiki/PPTools/Guiguts/Guiguts_2_Manual/Help_Menu#Regular_Expression_Quick_Reference"
        )

    def highlight_quotbrac_callback(self, value: bool) -> None:
        """Callback when highlight_quotbrac preference is changed."""
        if value:
            maintext().highlight_quotbrac()
        else:
            maintext().remove_highlights_quotbrac()

    def highlight_proofercomment_callback(self, value: bool) -> None:
        """Callback when highlight_proofercomment preference is changed."""
        if value:
            maintext().highlight_proofercomment()
        else:
            maintext().remove_highlights_proofercomment()

    def initialize_preferences(self) -> None:
        """Set default preferences and load settings from the GGPrefs file."""
        preferences.set_default(PrefKey.AUTO_IMAGE, False)
        preferences.set_callback(PrefKey.AUTO_IMAGE, self.auto_image_callback)
        preferences.set_default(PrefKey.BELL_AUDIBLE, True)
        preferences.set_default(PrefKey.BELL_VISUAL, True)
        preferences.set_default(PrefKey.IMAGE_WINDOW_DOCKED, True)
        preferences.set_default(PrefKey.RECENT_FILES, [])
        preferences.set_default(PrefKey.LINE_NUMBERS, True)
        preferences.set_default(PrefKey.ORDINAL_NAMES, True)
        preferences.set_callback(
            PrefKey.LINE_NUMBERS, lambda value: maintext().show_line_numbers(value)
        )
        preferences.set_default(PrefKey.SEARCH_HISTORY, [])
        preferences.set_default(PrefKey.REPLACE_HISTORY, [])
        preferences.set_default(PrefKey.SEARCHDIALOG_REVERSE, False)
        preferences.set_default(PrefKey.SEARCHDIALOG_MATCH_CASE, False)
        preferences.set_default(PrefKey.SEARCHDIALOG_WHOLE_WORD, False)
        preferences.set_default(PrefKey.SEARCHDIALOG_WRAP, True)
        preferences.set_default(PrefKey.SEARCHDIALOG_REGEX, False)
        preferences.set_default(PrefKey.SEARCHDIALOG_MULTI_REPLACE, False)
        preferences.set_default(PrefKey.DIALOG_GEOMETRY, {})
        preferences.set_default(PrefKey.ROOT_GEOMETRY, "800x400")
        preferences.set_default(PrefKey.ROOT_GEOMETRY_STATE, "normal")
        preferences.set_default(PrefKey.DEFAULT_LANGUAGES, "en")
        preferences.set_default(PrefKey.WFDIALOG_SUSPECTS_ONLY, False)
        preferences.set_default(PrefKey.WFDIALOG_IGNORE_CASE, False)
        preferences.set_default(PrefKey.WFDIALOG_DISPLAY_TYPE, WFDisplayType.ALL_WORDS)
        preferences.set_default(PrefKey.WFDIALOG_SORT_TYPE, WFSortType.ALPHABETIC)
        preferences.set_default(PrefKey.CHECKERDIALOG_SORT_TYPE_DICT, {})
        preferences.set_default(PrefKey.CHECKERDIALOG_SUSPECTS_ONLY_DICT, {})
        preferences.set_default(PrefKey.WFDIALOG_ITALIC_THRESHOLD, ["4"])
        preferences.set_default(PrefKey.WFDIALOG_REGEX, [])
        preferences.set_default(
            PrefKey.JEEBIES_PARANOIA_LEVEL, JeebiesParanoiaLevel.NORMAL
        )
        preferences.set_default(
            PrefKey.LEVENSHTEIN_DISTANCE, LevenshteinEditDistance.ONE
        )
        preferences.set_default(PrefKey.FOOTNOTE_INDEX_STYLE, FootnoteIndexStyle.NUMBER)
        preferences.set_default(PrefKey.WRAP_LEFT_MARGIN, 0)
        preferences.set_default(PrefKey.WRAP_RIGHT_MARGIN, 72)
        preferences.set_default(PrefKey.WRAP_BLOCKQUOTE_INDENT, 2)
        preferences.set_default(PrefKey.WRAP_BLOCKQUOTE_RIGHT_MARGIN, 72)
        preferences.set_default(PrefKey.WRAP_BLOCK_INDENT, 2)
        preferences.set_default(PrefKey.WRAP_POETRY_INDENT, 4)
        preferences.set_default(PrefKey.WRAP_INDEX_MAIN_MARGIN, 2)
        preferences.set_default(PrefKey.WRAP_INDEX_WRAP_MARGIN, 8)
        preferences.set_default(PrefKey.WRAP_INDEX_RIGHT_MARGIN, 72)
        preferences.set_default(PrefKey.TEXT_MARKUP_ITALIC, "_")
        preferences.set_default(PrefKey.TEXT_MARKUP_BOLD, "=")
        preferences.set_default(PrefKey.TEXT_MARKUP_SMALLCAPS, "+")
        preferences.set_default(PrefKey.TEXT_MARKUP_GESPERRT, "~")
        preferences.set_default(PrefKey.TEXT_MARKUP_FONT, "=")
        preferences.set_default(PrefKey.PAGESEP_AUTO_TYPE, PageSepAutoType.AUTO_FIX)
        preferences.set_default(PrefKey.THEME_NAME, "Default")
        preferences.set_callback(
            PrefKey.THEME_NAME,
            lambda value: themed_style().theme_use(
                theme_name_internal_from_user(value)
            ),
        )
        preferences.set_default(PrefKey.TEAROFF_MENUS, False)
        preferences.set_default(PrefKey.COMPOSE_HISTORY, [])
        # Since fonts aren't available until Tk has initialized, set default font family
        # to be empty string here, and set the true default later in MainText.
        preferences.set_default(PrefKey.TEXT_FONT_FAMILY, "")
        preferences.set_callback(
            PrefKey.TEXT_FONT_FAMILY,
            lambda *value: maintext().set_font(),
        )
        preferences.set_default(PrefKey.TEXT_FONT_SIZE, 12)
        preferences.set_callback(
            PrefKey.TEXT_FONT_SIZE,
            lambda *value: maintext().set_font(),
        )
        preferences.set_default(PrefKey.SPELL_THRESHOLD, 3)
        preferences.set_default(PrefKey.UNMATCHED_NESTABLE, False)
        preferences.set_default(
            PrefKey.UNICODE_BLOCK, UnicodeBlockDialog.commonly_used_characters_name
        )
        preferences.set_default(PrefKey.UNICODE_SEARCH_HISTORY, [])
        preferences.set_default(PrefKey.SPLIT_TEXT_WINDOW, False)
        preferences.set_default(PrefKey.SPLIT_TEXT_SASH_COORD, 0)
        preferences.set_default(PrefKey.IMAGE_INVERT, False)
        preferences.set_default(PrefKey.IMAGE_FLOAT_GEOMETRY, "400x600+100+100")
        preferences.set_default(PrefKey.IMAGE_DOCK_SASH_COORD, 300)
        preferences.set_default(PrefKey.IMAGE_SCALE_FACTOR, 0.5)
        preferences.set_default(PrefKey.IMAGE_VIEWER_ALERT, True)
        preferences.set_default(
            PrefKey.SCANNOS_FILENAME,
            str(DEFAULT_SCANNOS_DIR.joinpath(DEFAULT_REGEX_SCANNOS)),
        )
        preferences.set_default(
            PrefKey.SCANNOS_HISTORY,
            [
                str(DEFAULT_SCANNOS_DIR.joinpath(DEFAULT_REGEX_SCANNOS)),
                str(DEFAULT_SCANNOS_DIR.joinpath(DEFAULT_STEALTH_SCANNOS)),
                str(DEFAULT_SCANNOS_DIR.joinpath(DEFAULT_MISSPELLED_SCANNOS)),
            ],
        )
        preferences.set_default(PrefKey.HIGHLIGHT_QUOTBRAC, False)
        preferences.set_callback(
            PrefKey.HIGHLIGHT_QUOTBRAC, self.highlight_quotbrac_callback
        )
        preferences.set_default(PrefKey.COLUMN_NUMBERS, False)
        preferences.set_callback(
            PrefKey.COLUMN_NUMBERS, lambda value: maintext().show_column_numbers(value)
        )
        preferences.set_default(PrefKey.HTML_ITALIC_MARKUP, HTMLMarkupTypes.KEEP)
        preferences.set_default(PrefKey.HTML_BOLD_MARKUP, HTMLMarkupTypes.KEEP)
        preferences.set_default(PrefKey.HTML_GESPERRT_MARKUP, HTMLMarkupTypes.EM_CLASS)
        preferences.set_default(PrefKey.HTML_FONT_MARKUP, HTMLMarkupTypes.SPAN_CLASS)
        preferences.set_default(
            PrefKey.HTML_UNDERLINE_MARKUP, HTMLMarkupTypes.SPAN_CLASS
        )
        preferences.set_default(PrefKey.HTML_SHOW_PAGE_NUMBERS, True)
        preferences.set_default(PrefKey.HTML_IMAGE_UNIT, "%")
        preferences.set_default(PrefKey.HTML_IMAGE_OVERRIDE_EPUB, True)
        preferences.set_default(PrefKey.HTML_IMAGE_ALIGNMENT, "center")
        preferences.set_default(PrefKey.HIGHLIGHT_PROOFERCOMMENT, True)
        preferences.set_callback(
            PrefKey.HIGHLIGHT_PROOFERCOMMENT, self.highlight_proofercomment_callback
        )

        # Check all preferences have a default
        for pref_key in PrefKey:
            assert preferences.get_default(pref_key) is not None

        # If `--nohome` argument given, Prefs are not loaded & saved in Prefs file
        preferences.set_permanent(not self.args.nohome)
        preferences.load()

    # Lay out menus
    def init_menus(self) -> None:
        """Create all the menus."""
        self.init_file_menu()
        self.init_edit_menu()
        self.init_search_menu()
        self.init_tools_menu()
        self.init_text_menu()
        self.init_html_menu()
        self.init_view_menu()
        self.init_help_menu()
        self.init_os_menu()

        if is_mac():
            root().createcommand(
                "tk::mac::ShowPreferences", PreferencesDialog.show_dialog
            )
            root().createcommand("tk::mac::OpenDocument", self.open_document)
            root().createcommand("tk::mac::Quit", self.quit_program)

    def init_file_menu(self) -> None:
        """(Re-)create the File menu."""
        try:
            self.file_menu.delete(0, "end")  # type:ignore[union-attr]
        except AttributeError:
            self.file_menu = Menu(menubar(), "~File")
        assert self.file_menu is not None
        self.file_menu.add_button("~Open...", self.open_file, "Cmd/Ctrl+O")
        self.init_file_recent_menu(self.file_menu)
        self.file_menu.add_button("~Save", self.file.save_file, "Cmd/Ctrl+S")
        self.file_menu.add_button(
            "Save ~As...", self.file.save_as_file, "Cmd/Ctrl+Shift+S"
        )
        self.file_menu.add_button("Sa~ve a Copy As...", self.file.save_copy_as_file)
        self.file_menu.add_button(
            "~Close", self.close_command, "Cmd+W" if is_mac() else ""
        )
        self.init_file_project_menu(self.file_menu)
        if not is_mac():
            self.file_menu.add_separator()
            self.file_menu.add_button("E~xit", self.quit_program, "")

    def init_file_recent_menu(self, parent: Menu) -> None:
        """Create the Recent Documents menu."""
        recent_menu = Menu(parent, "Recent Doc~uments")
        for count, file in enumerate(preferences.get(PrefKey.RECENT_FILES), start=1):
            recent_menu.add_button(
                f"~{count}: {file}",
                lambda fn=file: self.open_file(fn),  # type:ignore[misc]
            )

    def init_file_project_menu(self, parent: Menu) -> None:
        """Create the File->Project menu."""
        project_menu = Menu(parent, "~Project")
        project_menu.add_button(
            "Add ~Good/Bad Words to Project Dictionary",
            self.file.add_good_and_bad_words,
        )
        project_menu.add_button(
            f"Set ~Scan Image {folder_dir_str()}...",
            self.file.choose_image_dir,
        )
        project_menu.add_button("Set ~Language(s)...", self.file.set_languages)
        project_menu.add_separator()
        project_menu.add_button(
            "~Configure Page Labels...", self.show_page_details_dialog
        )
        project_menu.add_separator()
        project_menu.add_button("~Add Page Marker Flags", self.file.add_page_flags)
        project_menu.add_button(
            "~Remove Page Marker Flags", self.file.remove_page_flags
        )

    def init_edit_menu(self) -> None:
        """Create the Edit menu."""
        edit_menu = Menu(menubar(), "~Edit")
        edit_menu.add_button("~Undo", "<<Undo>>", "Cmd/Ctrl+Z")
        edit_menu.add_button(
            "~Redo", "<<Redo>>", "Cmd+Shift+Z" if is_mac() else "Ctrl+Y"
        )
        edit_menu.add_separator()
        edit_menu.add_cut_copy_paste()
        edit_menu.add_button(
            "R~estore Selection",
            maintext().restore_selection_ranges,
        )
        edit_menu.add_separator()
        edit_menu.add_button(
            "Co~lumn Cut", maintext().columnize_cut, "Cmd/Ctrl+Shift+X"
        )
        edit_menu.add_button(
            "C~olumn Copy", maintext().columnize_copy, "Cmd/Ctrl+Shift+C"
        )
        edit_menu.add_button(
            "Colu~mn Paste", maintext().columnize_paste, "Cmd/Ctrl+Shift+V"
        )
        edit_menu.add_button(
            "To~ggle Column/Regular Selection",
            maintext().toggle_selection_type,
        )
        edit_menu.add_separator()
        case_menu = Menu(edit_menu, "C~hange Case")
        case_menu.add_button(
            "~lowercase selection",
            lambda: maintext().transform_selection(str.lower),
            "",
        )
        case_menu.add_button(
            "~Sentence case selection",
            lambda: maintext().transform_selection(
                maintext().sentence_case_transformer
            ),
            "",
        )
        case_menu.add_button(
            "~Title Case Selection",
            lambda: maintext().transform_selection(maintext().title_case_transformer),
            "",
        )
        case_menu.add_button(
            "~UPPERCASE SELECTION",
            lambda: maintext().transform_selection(str.upper),
            "",
        )
        if not is_mac():
            edit_menu.add_separator()
            edit_menu.add_button("Pre~ferences...", PreferencesDialog.show_dialog)

    def init_search_menu(self) -> None:
        """Create the Search menu."""
        search_menu = Menu(menubar(), "~Search")
        search_menu.add_button(
            "~Search & Replace...",
            show_search_dialog,
            "Cmd/Ctrl+F",
        )
        search_menu.add_button(
            "Find Ne~xt",
            find_next,
            "Cmd+G" if is_mac() else "F3",
        )
        search_menu.add_button(
            "Find Pre~vious",
            lambda: find_next(backwards=True),
            "Cmd+Shift+G" if is_mac() else "Shift+F3",
        )
        search_menu.add_separator()
        self.init_search_goto_menu(search_menu)
        search_menu.add_separator()
        search_menu.add_button(
            "Find Proofer ~Comments",
            proofer_comment_check,
        )
        search_menu.add_button(
            "Find ~Asterisks w/o Slash",
            asterisk_check,
        )
        search_menu.add_separator()
        search_menu.add_button(
            "Highlight Single ~Quotes in Selection",
            maintext().highlight_single_quotes,
        )
        search_menu.add_button(
            "Highlight ~Double Quotes in Selection",
            maintext().highlight_double_quotes,
        )
        search_menu.add_button(
            "~Remove Highlights",
            maintext().remove_highlights,
        )
        search_menu.add_separator()
        search_menu.add_checkbox(
            "Highlight S~urrounding Quotes & Brackets",
            root().highlight_quotbrac,
        )
        search_menu.add_checkbox(
            "Highlight Al~ignment Column",
            maintext().aligncol_active,
            lambda: maintext().highlight_aligncol_callback(True),
            lambda: maintext().highlight_aligncol_callback(False),
        )
        search_menu.add_checkbox(
            "Highlight ~Proofer Comments",
            root().highlight_proofercomment,
        )
        search_menu.add_separator()
        self.init_bookmark_menu(search_menu)

    def init_search_goto_menu(self, parent: Menu) -> None:
        """Create the Search->Goto menu."""
        goto_menu = Menu(parent, "~Goto")
        goto_menu.add_button(
            "~Line Number...",
            self.file.goto_line,
        )
        goto_menu.add_button(
            "Pa~ge (png)...",
            self.file.goto_image,
        )
        goto_menu.add_button(
            "Page La~bel...",
            self.file.goto_page,
        )
        goto_menu.add_button(
            "~Previous Page",
            self.file.prev_page,
        )
        goto_menu.add_button(
            "~Next Page",
            self.file.next_page,
        )

    def init_bookmark_menu(self, parent: Menu) -> None:
        """Create the Bookmarks menu."""
        bookmark_menu = Menu(parent, "~Bookmarks")
        # Because keyboard layouts are different, need to bind to several keys for some bookmarks
        shortcuts = [
            ("exclam",),
            ("at", "quotedbl"),
            ("numbersign", "sterling", "section", "periodcentered"),
            ("dollar", "currency"),
            ("percent",),
        ]
        # Shift+Cmd+number no good for Mac due to clash with screen capture shortcuts
        # Ctrl+number could also clash on Mac with virtual desktop switching, but
        # the best option we have that works at the moment.
        modifier = "Shift"
        for bm, keys in enumerate(shortcuts, start=1):
            bookmark_menu.add_button(
                f"Set Bookmark {bm}",
                lambda num=bm: self.file.set_bookmark(num),  # type:ignore[misc]
                f"{modifier}+Ctrl+Key-{bm}",
            )
            # Add extra shortcuts to cope with keyboard layout differences
            for key in keys:
                (_, key_event) = process_accel(f"{modifier}+Ctrl+{key}")
                maintext().key_bind(
                    key_event,
                    lambda _event, num=bm: self.file.set_bookmark(  # type:ignore[misc]
                        num
                    ),
                    bind_all=True,
                )

        for bm in range(1, 6):
            bookmark_menu.add_button(
                f"Go To Bookmark {bm}",
                lambda num=bm: self.file.goto_bookmark(num),  # type:ignore[misc]
                f"Ctrl+Key-{bm}",
            )

    def init_tools_menu(self) -> None:
        """Create the Tools menu."""
        tools_menu = Menu(menubar(), "~Tools")
        tools_menu.add_button("Basic Fi~xup...", basic_fixup_check)
        tools_menu.add_button("~Word Frequency...", word_frequency)
        tools_menu.add_button("~Bookloupe...", bookloupe_check)
        tools_menu.add_button(
            "~Spelling...",
            lambda: spell_check(
                self.file.project_dict,
                self.file.add_good_word_to_project_dictionary,
                self.file.add_good_word_to_global_dictionary,
            ),
        )
        tools_menu.add_button("~Jeebies...", jeebies_check)
        tools_menu.add_button("Stealth S~cannos...", stealth_scannos)
        tools_menu.add_button(
            "Word ~Distance Check...", lambda: levenshtein_check(self.file.project_dict)
        )
        tools_menu.add_separator()
        tools_menu.add_button("~Page Separator Fixup...", page_separator_fixup)
        tools_menu.add_button("~Footnote Fixup...", footnote_check)
        tools_menu.add_button("Side~note Fixup...", lambda: illosn_check("Sidenote"))
        tools_menu.add_button(
            "~Illustration Fixup...", lambda: illosn_check("Illustration")
        )
        tools_menu.add_separator()
        tools_menu.add_button("~Rewrap All", self.file.rewrap_all)
        tools_menu.add_button("R~ewrap Selection", self.file.rewrap_selection)
        tools_menu.add_button("C~lean Up Rewrap Markers", self.file.rewrap_cleanup)
        tools_menu.add_separator()
        tools_menu.add_button("Convert to Curly ~Quotes", convert_to_curly_quotes)
        tools_menu.add_button("Check Curly Q~uotes", check_curly_quotes)
        unmatched_menu = Menu(tools_menu, "Un~matched")
        unmatched_menu.add_button("Bloc~k Markup...", unmatched_block_markup)
        unmatched_menu.add_button("~DP Markup...", unmatched_dp_markup)
        unmatched_menu.add_button("~Brackets...", unmatched_brackets)
        unmatched_menu.add_button("Curly ~Quotes...", unmatched_curly_quotes)

        fraction_menu = Menu(tools_menu, "~Convert Fractions")
        fraction_menu.add_button(
            "~Unicode Fractions Only",
            lambda: fraction_convert(FractionConvertType.UNICODE),
        )
        fraction_menu.add_button(
            "Unicode Fractions ~Or Super/Subscript",
            lambda: fraction_convert(FractionConvertType.MIXED),
        )
        fraction_menu.add_button(
            "All to ~Super/Subscript",
            lambda: fraction_convert(FractionConvertType.SUPSUB),
        )
        unicode_menu = Menu(tools_menu, "~Unicode")
        unicode_menu.add_button(
            "Unicode ~Search/Entry", UnicodeSearchDialog.show_dialog
        )
        unicode_menu.add_button("Unicode ~Blocks", UnicodeBlockDialog.show_dialog)
        unicode_menu.add_button(
            "~Normalize Selected Characters",
            unicode_normalize,
        )
        unicode_menu.add_button(
            "~Compose Sequence...",
            ComposeSequenceDialog.show_dialog,
            "Cmd/Ctrl+I",
        )

    def init_text_menu(self) -> None:
        """Create the Text menu."""
        text_menu = Menu(menubar(), "Te~xt")
        text_menu.add_button(
            "Convert ~Markup...", TextMarkupConvertorDialog.show_dialog
        )
        text_menu.add_button("PP~txt...", lambda: pptxt(self.file.project_dict))

    def init_html_menu(self) -> None:
        """Create the HTML menu."""
        html_menu = Menu(menubar(), "HT~ML")
        html_menu.add_button("HTML ~Generator...", HTMLGeneratorDialog.show_dialog)
        html_menu.add_button("Auto-~Illustrations...", HTMLImageDialog.show_dialog)
        html_menu.add_button("~Unmatched HTML Tags...", unmatched_html_markup)
        html_menu.add_button("HTML5 ~Validator (online)...", html_validator_check)
        html_menu.add_button("HTML ~Link Checker...", html_link_check)

    def init_view_menu(self) -> None:
        """Create the View menu."""
        view_menu = Menu(menubar(), "~View")
        view_menu.add_checkbox(
            "Split ~Text Window",
            root().split_text_window,
            lambda: maintext().show_peer(),
            lambda: maintext().hide_peer(),
        )
        view_menu.add_checkbox(
            "~Dock Image",
            root().image_window_docked_state,
            self.mainwindow.dock_image,
            self.mainwindow.float_image,
        )
        view_menu.add_checkbox(
            "~Auto Image",
            root().auto_image_state,
        )
        view_menu.add_button("~See Image", self.show_image)
        view_menu.add_button("~Hide Image", self.hide_image)
        view_menu.add_checkbox(
            "~Invert Image",
            root().invert_image_state,
            mainimage().show_image,
            mainimage().show_image,
        )
        view_menu.add_separator()
        view_menu.add_button("~Message Log", self.mainwindow.messagelog.show)
        view_menu.add_separator()
        if not is_mac():  # Full Screen behaves oddly on Macs
            view_menu.add_checkbox(
                "~Full Screen",
                root().full_screen_var,
                lambda: root().wm_attributes("-fullscreen", True),
                lambda: root().wm_attributes("-fullscreen", False),
            )

    def init_help_menu(self) -> None:
        """Create the Help menu."""
        help_menu = Menu(menubar(), "~Help")
        help_menu.add_button(
            "Guiguts ~Manual", self.show_help_manual, "F1", force_main_only=True
        )
        help_menu.add_button("About ~Guiguts", self.help_about)
        help_menu.add_button("~Regex Quick Reference", self.show_help_regex)
        help_menu.add_button(
            "List of ~Compose Sequences", ComposeHelpDialog.show_dialog
        )

    def init_os_menu(self) -> None:
        """Create the OS-specific menu.

        Currently only does anything on Macs
        """
        if is_mac():
            # Window menu
            Menu(menubar(), "Window", name="window")

    def init_statusbar(self, the_statusbar: StatusBar) -> None:
        """Add labels to initialize the statusbar.

        Functions are bound to Button Release, rather than Button Press, since
        that is how buttons normally operate.
        """

        def rowcol_str() -> str:
            """Format current insert index for statusbar."""
            row, col = maintext().get_insert_index().rowcol()
            return f"L:{row} C:{col}"

        the_statusbar.add(
            "rowcol",
            tooltip="Click: Go to line\nShift click: Toggle line numbers\nShift right-click: Toggle column numbers",
            update=rowcol_str,
        )
        the_statusbar.add_binding("rowcol", "ButtonRelease-1", self.file.goto_line)
        the_statusbar.add_binding(
            "rowcol",
            "Shift-ButtonRelease-1",
            lambda: preferences.toggle(PrefKey.LINE_NUMBERS),
        )
        the_statusbar.add_binding(
            "rowcol",
            "Shift-ButtonRelease-3",
            lambda: preferences.toggle(PrefKey.COLUMN_NUMBERS),
        )

        the_statusbar.add(
            "img",
            tooltip="Click: Go to image",
            update=lambda: "Img: " + maintext().get_current_image_name(),
        )
        the_statusbar.add_binding("img", "ButtonRelease-1", self.file.goto_image)

        the_statusbar.add(
            "prev img", tooltip="Click: Go to previous image", text="<", width=1
        )
        the_statusbar.add_binding("prev img", "ButtonRelease-1", self.file.prev_page)

        the_statusbar.add(
            "see img",
            tooltip="Click: See image\nShift click: Toggle auto-image",
            text="See Img",
            width=9,
        )
        the_statusbar.add_binding(
            "see img",
            "ButtonRelease-1",
            self.show_image,
        )
        the_statusbar.add_binding(
            "see img",
            "Shift-ButtonRelease-1",
            lambda: preferences.toggle(PrefKey.AUTO_IMAGE),
        )

        the_statusbar.add(
            "next img", tooltip="Click: Go to next image", text=">", width=1
        )
        the_statusbar.add_binding("next img", "ButtonRelease-1", self.file.next_page)

        the_statusbar.add(
            "page label",
            tooltip="Click: Go to page\nShift click: Configure page labels",
            text="Lbl: ",
            update=lambda: "Lbl: " + self.file.get_current_page_label(),
        )
        the_statusbar.add_binding("page label", "ButtonRelease-1", self.file.goto_page)
        the_statusbar.add_binding(
            "page label", "Shift-ButtonRelease-1", self.show_page_details_dialog
        )

        def selection_str() -> str:
            """Format current selection range for statusbar.

            Returns:
                "Start-End" for a regular selection, and "R:rows C:cols" for column selection.
            """
            maintext().selection_cursor()
            ranges = maintext().selected_ranges()
            maintext().save_selection_ranges()
            if not ranges:
                return "No selection"
            if len(ranges) == 1:
                return f"{ranges[0].start.index()}-{ranges[-1].end.index()}"
            row_diff = abs(ranges[-1].end.row - ranges[0].start.row) + 1
            col_diff = abs(ranges[-1].end.col - ranges[0].start.col)
            return f"R:{row_diff} C:{col_diff}"

        the_statusbar.add(
            "selection",
            tooltip="Click: Restore previous selection\nShift Click: Toggle column/regular selection",
            update=selection_str,
            width=16,
        )
        the_statusbar.add_binding(
            "selection", "ButtonRelease-1", maintext().restore_selection_ranges
        )
        the_statusbar.add_binding(
            "selection", "Shift-ButtonRelease-1", maintext().toggle_selection_type
        )

        the_statusbar.add(
            "languages label",
            tooltip="Click: Set language(s), e.g. 'en' or 'de fr'",
            text="Lang: ",
        )
        the_statusbar.add_binding(
            "languages label", "ButtonRelease-1", self.file.set_languages
        )

        def ordinal_name_display(force: Optional[bool] = None) -> None:
            """Set, unset, or toggle, name display in ordinal button.

            Args:
                Force: True to set, False to unset, omit to toggle.
            """
            if force is None:
                force = not preferences.get(PrefKey.ORDINAL_NAMES)
            preferences.set(PrefKey.ORDINAL_NAMES, force)

        def ordinal_str() -> str:
            """Format ordinal of single selected char, or char at current insert
            index for statusbar."""
            # Get character - display nothing if more than one char selected
            sel_ranges = maintext().selected_ranges()
            if len(sel_ranges) == 0:
                char = maintext().get(maintext().get_insert_index().index())
            elif len(sel_ranges) == 1:
                char = maintext().selected_text()
                if len(char) != 1:
                    return ""
            else:
                return ""
            # unicodedata.name fails to return name for "control" characters
            # but the only one we care about is line feed
            if preferences.get(PrefKey.ORDINAL_NAMES):
                try:
                    name = f": {unicodedata.name(char)}"
                except ValueError:
                    name = ": LINE FEED" if char == "\n" else ""
            else:
                name = ""
            return f"U+{ord(char):04x}{name}"

        the_statusbar.add(
            "ordinal",
            tooltip="Character after the cursor\nClick to toggle name display",
            update=ordinal_str,
        )
        the_statusbar.add_binding("ordinal", "ButtonRelease-1", ordinal_name_display)

    def logging_init(self) -> None:
        """Set up basic logger until GUI is ready."""
        if self.args.debug:
            log_level = logging.DEBUG
            console_log_level = logging.DEBUG
            formatter = logging.Formatter(DEBUG_FORMAT, "%H:%M:%S")
        else:
            log_level = logging.INFO
            console_log_level = logging.WARNING
            formatter = logging.Formatter(MESSAGE_FORMAT, "%H:%M:%S")
        logger.setLevel(log_level)
        # Output to console
        console_handler = logging.StreamHandler()
        console_handler.setLevel(console_log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    def logging_add_gui(self) -> None:
        """Add handlers to display log messages via the GUI.

        Assumes mainwindow has created the message_log handler.
        """

        # Message log is approximate GUI equivalent to console output
        if self.args.debug:
            message_log_level = logging.DEBUG
            formatter = logging.Formatter(DEBUG_FORMAT, "%H:%M:%S")
        else:
            message_log_level = logging.INFO
            formatter = logging.Formatter(MESSAGE_FORMAT, "%H:%M:%S")
        self.mainwindow.messagelog.setLevel(message_log_level)
        self.mainwindow.messagelog.setFormatter(formatter)
        logger.addHandler(self.mainwindow.messagelog)

        # Alert is just for errors, e.g. unable to load file
        alert_handler = ErrorHandler()
        alert_handler.setLevel(logging.ERROR)
        alert_handler.setFormatter(formatter)
        logger.addHandler(alert_handler)


def main() -> None:
    """Main application function."""
    Guiguts().run()


if __name__ == "__main__":
    main()



"""Functions to convert from text to HTML."""

from enum import StrEnum, auto
import importlib.resources
import logging
from pathlib import Path
import tkinter as tk
from tkinter import ttk, messagebox
from typing import Optional

import regex as re

from guiguts.data import html
from guiguts.file import the_file
from guiguts.maintext import maintext
from guiguts.page_details import PAGE_LABEL_PREFIX, PageDetail
from guiguts.preferences import (
    preferences,
    PrefKey,
    PersistentString,
    PersistentBoolean,
)
from guiguts.utilities import IndexRange, DiacriticRemover, IndexRowCol
from guiguts.widgets import ToplevelDialog, Busy

logger = logging.getLogger(__package__)

css_indents: set[int] = set()
book_title: Optional[str] = None
DEFAULT_HTML_DIR = importlib.resources.files(html)
HTML_HEADER_NAME = "html_header.txt"
PAGE_ID_PREFIX = "Page_"
H1_SKIP_REGEX = r"(\[Illustr|/[*$fxcr]|[*$fxcr]/)"

inline_conversion_dict = {
    PrefKey.HTML_ITALIC_MARKUP: ("i", "italic"),
    PrefKey.HTML_BOLD_MARKUP: ("b", "bold"),
    PrefKey.HTML_GESPERRT_MARKUP: ("g", "gesperrt"),
    PrefKey.HTML_FONT_MARKUP: ("f", "antiqua"),
    PrefKey.HTML_UNDERLINE_MARKUP: ("u", "u"),
}


class HTMLMarkupTypes(StrEnum):
    """Enum class to store values for markup conversion."""

    KEEP = auto()
    EM = auto()
    EM_CLASS = auto()
    SPAN_CLASS = auto()


class HTMLGeneratorDialog(ToplevelDialog):
    """Dialog for converting text file to HTML."""

    manual_page = "HTML_Menu#HTML_Generator"
    book_title = None

    def __init__(self) -> None:
        """Initialize HTML Generator dialog."""
        super().__init__("HTML Auto-generator", resize_x=False, resize_y=False)

        # Title
        title_frame = ttk.Frame(self.top_frame, padding=2)
        title_frame.grid(row=0, column=0, sticky="NSEW")
        title_frame.columnconfigure(1, weight=1)
        ttk.Label(title_frame, text="Title:").grid(row=0, column=0, padx=(0, 5))
        HTMLGeneratorDialog.book_title = tk.StringVar(self, get_title())
        self.title_entry = ttk.Entry(
            title_frame,
            textvariable=HTMLGeneratorDialog.book_title,
        )
        self.title_entry.grid(row=0, column=1, sticky="NSEW")
        self.title_entry.focus_set()

        # Whether to display page numbers
        self.display_page_numbers = ttk.Checkbutton(
            self.top_frame,
            text="Show Page Numbers in HTML",
            variable=PersistentBoolean(PrefKey.HTML_SHOW_PAGE_NUMBERS),
            takefocus=False,
        )
        self.display_page_numbers.grid(row=1, column=0, sticky="NSEW", pady=5)

        # Markup conversion
        markup_frame = ttk.LabelFrame(self.top_frame, text="Inline Markup", padding=2)
        markup_frame.grid(row=2, column=0, sticky="NSEW")
        for col, text in enumerate(
            ("Keep", "<em>", "<em class>", "<span class>"), start=1
        ):
            ttk.Label(markup_frame, text=text, anchor=tk.CENTER).grid(
                row=0, column=col, padx=5
            )
        for row, (key, (letter, _)) in enumerate(
            inline_conversion_dict.items(), start=1
        ):
            ttk.Label(markup_frame, text=f"<{letter}>:").grid(
                row=row, column=0, sticky="NSE", padx=(0, 5)
            )
            type_var = PersistentString(key)
            for col, value in enumerate(
                (
                    HTMLMarkupTypes.KEEP,
                    HTMLMarkupTypes.EM,
                    HTMLMarkupTypes.EM_CLASS,
                    HTMLMarkupTypes.SPAN_CLASS,
                ),
                start=1,
            ):
                ttk.Radiobutton(
                    markup_frame,
                    text="",
                    variable=type_var,
                    value=value,
                    takefocus=False,
                ).grid(row=row, column=col, padx=(7, 0))

        ttk.Button(
            self.top_frame,
            text="Auto-generate HTML",
            command=html_autogenerate,
            takefocus=False,
        ).grid(row=3, column=0, pady=2)


def html_autogenerate() -> None:
    """Autogenerate HTML from text file."""
    Busy.busy()
    backup_fn = re.sub(r"\.[^\.]*$", "-htmlbak.txt", the_file().filename)
    the_file().save_copy(backup_fn)
    try:
        do_html_autogenerate()
    except SyntaxError as exc:
        logger.error(exc)
        # Check if user would like to reload backup file
        reload = messagebox.askokcancel(
            title="Re-load backup file?",
            message="An error occurred during HTML autogeneration,\nand the file was not completely converted.",
            detail='Click "OK" to re-load pre-conversion backup file,\nor "Cancel" to explore partially converted file (see manual).',
            icon=messagebox.QUESTION,
        )
        if reload:
            # Save name because load_file will overwrite it
            old_filename = the_file().filename
            the_file().load_file(backup_fn)
            # Restore original filename
            the_file().filename = old_filename
            # Put it at the top of the recent files list
            the_file().store_recent_file(old_filename)
        # Go to error line number if available
        line_num, is_line_num = re.subn(r"Line (\d+):.*", r"\1", str(exc))
        if is_line_num:
            maintext().set_insert_index(IndexRowCol(int(line_num), 0))

    Busy.unbusy()


def do_html_autogenerate() -> None:
    """Do the work of HTML autogenerate."""
    css_indents.clear()
    maintext().undo_block_begin()
    remove_trailing_spaces()
    adjust_pagemark_postions()
    html_convert_entities()
    html_convert_title()
    html_convert_body()
    html_convert_inline()
    html_convert_smallcaps()
    html_convert_footnotes()
    html_convert_page_anchors()
    html_convert_footnote_landing_zones()
    html_convert_sidenotes()
    html_add_chapter_divs()
    html_wrap_long_lines()
    maintext().set_insert_index(maintext().start())


def remove_trailing_spaces() -> None:
    """Remove trailing spaces from every line."""
    maintext().replace_all(" +$", "", regexp=True)


def html_convert_entities() -> None:
    """Replace occurrences of ampersand, less & greater than, non-breaking space
    with HTML entities (except for things like `<i>`)."""
    maintext().replace_all("&", "&amp;")
    maintext().replace_all("\xa0", "&nbsp;")
    maintext().replace_all("--", "—")

    # Find all < and > characters
    search_range = maintext().start_to_end()
    while match := maintext().find_match("(<|>)", search_range, regexp=True):
        match_index = match.rowcol.index()
        match_text = maintext().get(match_index, f"{match_index} lineend")
        # Check if it's a DP tag, like `<i>` or `</sc>`
        check_match = re.match("</?(i|b|f|g|u|sc|tb)>", match_text)
        if check_match is None:
            maintext().replace(
                match_index,
                f"{match_index}+1c",
                "&lt;" if match_text[0] == "<" else "&gt;",
            )
            advance = 4
        else:
            advance = len(check_match[0])  # Skip whole of DP tag
        search_range = IndexRange(
            maintext().rowcol(f"{match_index}+{advance}c"), maintext().end()
        )


def html_convert_title() -> None:
    """Find first bit of text that may be the book title, and mark it up with h1."""
    in_title = False
    for step in range(1, 500):
        if maintext().compare(f"{step}.0", ">=", tk.END):
            break
        selection = maintext().get(f"{step}.0", f"{step}.end")
        if in_title:
            # Check if title complete (blank line or end of frontmatter/center/right/etc markup)
            if not selection:
                maintext().insert(f"{step - 1}.end", "</h1>")
                return
            if re.match("[*$fxcr]/", selection, flags=re.IGNORECASE):
                maintext().replace(f"{step}.0", f"{step}.end", "</h1>")
                return
            # Not complete - add <br> to end of previous line to preserve line breaks
            maintext().insert(f"{step - 1}.end", "<br>")
            continue
        # Not in title yet, so skip blank lines, illos or block markup
        if not selection or re.match(H1_SKIP_REGEX, selection, flags=re.IGNORECASE):
            continue
        # Found start of title
        in_title = True
        if step > 1 and re.match(
            "/[*$fxcr]",
            maintext().get(f"{step - 1}.0", f"{step - 1}.end"),
        ):
            maintext().replace(f"{step - 1}.0", f"{step - 1}.end", "<h1>")
        else:
            maintext().insert(f"{step}.0", "<h1>")


def html_convert_inline() -> None:
    """Replace occurrences of <i>...</i> and other inline markup
    with em/span depending on settings."""
    for key, (letter, classname) in inline_conversion_dict.items():
        match preferences.get(key):
            case HTMLMarkupTypes.EM:
                maintext().replace_all(f"<{letter}>", "<em>")
                maintext().replace_all(f"</{letter}>", "</em>")
            case HTMLMarkupTypes.EM_CLASS:
                maintext().replace_all(f"<{letter}>", f'<em class="{classname}">')
                maintext().replace_all(f"</{letter}>", "</em>")
            case HTMLMarkupTypes.SPAN_CLASS:
                maintext().replace_all(f"<{letter}>", f'<span class="{classname}">')
                maintext().replace_all(f"</{letter}>", "</span>")


def html_convert_smallcaps() -> None:
    """Replace occurrences of <sc>...</sc> with HTML span.

    Classname is `smcap` unless string contains no lowercase when it is `allsmcap`
    """
    # Easier to work backward so replacements don't affect search for next occurrence
    search_index = "end"
    while smcap_end := maintext().search("</sc>", search_index, backwards=True):
        smcap_start = maintext().search("<sc>", smcap_end, backwards=True)
        if not smcap_start:
            return
        test_text = maintext().get(f"{smcap_start}+4c", smcap_end)
        classname = (
            "smcap" if re.search(r"\p{Lowercase_Letter}", test_text) else "allsmcap"
        )
        maintext().replace(smcap_end, f"{smcap_end}+5c", "</span>")
        maintext().replace(
            smcap_start, f"{smcap_start}+4c", f'<span class="{classname}">'
        )


def html_convert_body() -> None:
    """Convert body of text to HTML one line at a time."""
    next_step = 1
    contents_start = "1.0"
    in_chap_heading = False
    in_h1_heading = False
    chap_id = ""
    chap_heading = ""
    auto_toc = ""
    in_para = False
    in_front_para = False
    in_stanza = False
    pre_flag = False
    front_flag = False
    poetry_flag = False
    list_flag = False
    index_flag = False
    dollar_nowrap_flag = False
    asterisk_nowrap_flag = False
    center_nowrap_flag = False
    right_nowrap_flag = False
    right_block_line_num = 0
    poetry_indent = 0
    blockquote_level = 0
    right_line_lengths: list[int] = []
    ibs_dict = {
        "i": False,
        "b": False,
        "sc": False,
    }  # Flags to track open inline markup across lines

    def check_illegal_nesting() -> None:
        """Raise exception if already inside block markup."""
        if (
            pre_flag  # pylint: disable=too-many-boolean-expressions
            or front_flag
            or poetry_flag
            or list_flag
            or dollar_nowrap_flag
            or asterisk_nowrap_flag
            or center_nowrap_flag
            or right_nowrap_flag
            or index_flag
        ):
            raise SyntaxError(f"Line {step}: Illegally nested block markup")

    def reset_ibs_dict() -> None:
        """Set the italic/bold/smcap flags to False when starting a new block."""
        for key in ibs_dict:
            ibs_dict[key] = False

    markup_start = 1
    chap_start = 1
    while next_step <= maintext().end().row:
        step = next_step
        next_step += 1
        line_start = f"{step}.0"
        line_end = f"{step}.end"
        selection = maintext().get(line_start, line_end)
        n_spaces = len(selection) - len(selection.lstrip())

        # Remove trailing spaces first
        sel_strip = selection.rstrip()
        if n_trail := len(selection) - len(sel_strip):
            maintext().delete(f"{line_end}-{n_trail}c", line_end)
        selection = sel_strip
        selection_lower = selection.lower()

        # Note start of table of contents (English-only atm)
        if step < 100 and contents_start == "1.0" and "contents" in selection_lower:
            contents_start = f"{line_start}+2l"

        selection = html_convert_sub_super(selection, line_start, line_end)
        if html_convert_tb(selection, line_start, line_end):
            continue

        # Skip h1 heading - dealt with earlier
        if "<h1>" in selection_lower:
            in_h1_heading = True
        if in_h1_heading:
            if "</h1>" in selection_lower:
                in_h1_heading = False
            continue

        # "/x" is replaced with "<pre>", then leave lines unchanged until "/x"-->"</pre>"
        if selection_lower == "/x":  # open
            maintext().replace(line_start, line_end, "<pre>")
            pre_flag = True
            markup_start = step
            continue
        if pre_flag:
            if selection_lower == "x/":  # close
                maintext().replace(line_start, line_end, "</pre>")
                pre_flag = False
            continue

        # Remove leading spaces now "pre" is dealt with.
        # n_spaces stores the number of spaces if needed later.
        maintext().delete(line_start, f"{line_start}+{n_spaces}c")

        # "/f" --> center all paragraphs until we get "f/"
        if selection_lower == "/f":  # open
            maintext().delete(line_start, line_end)
            front_flag = True
            markup_start = step
            in_front_para = False
            continue
        if front_flag:
            if selection_lower == "f/":  # close
                if in_front_para:
                    maintext().insert(f"{line_start} -1l lineend", "</p>")
                maintext().delete(line_start, line_end)
                front_flag = False
            elif selection:
                if not in_front_para:  # start of paragraph
                    maintext().insert(line_start, '<p class="center">')
                    in_front_para = True
            elif in_front_para:  # blank line - end of paragraph
                maintext().insert(f"{line_start} -1l lineend", "</p>")
                in_front_para = False
            continue

        # "/p" --> poetry until we get "p/"
        if selection_lower == "/p":  # open
            maintext().replace(
                line_start,
                line_end,
                '<div class="poetry-container"><div class="poetry">',
            )
            poetry_flag = True
            markup_start = step
            in_stanza = False
            poetry_end = maintext().search(
                "^p/$", line_end, tk.END, regexp=True, nocase=True
            )
            if not poetry_end:
                raise SyntaxError(f"Line {step}: Unclosed poetry")
            poetry_indent = poetry_indentation(
                maintext().get(f"{line_start}+1l", poetry_end)
            )
            reset_ibs_dict()
            continue
        if poetry_flag:
            if selection_lower == "p/":  # close
                if in_stanza:
                    maintext().insert(f"{line_start} -1l lineend", "</div>")
                maintext().replace(line_start, line_end, "</div></div>")
                poetry_flag = False
            elif selection:
                # Handle line numbers (at least 2 spaces plus digits at end of line)
                # Remove line number prior to handling per-line bold/ital/smcap
                # then add marked up line number afterwards.
                replacement = ""
                if linenum_match := re.search(r" {2,}(\d+) *$", selection):
                    linenum_len = len(linenum_match[0])
                    replacement = f'<span class="linenum">{linenum_match[1]}</span>'
                    maintext().delete(f"{line_end}-{linenum_len}c", line_end)
                do_per_line_markup(selection, line_start, line_end, ibs_dict)
                # Now safe to add indent markup and line number markup
                indent = n_spaces - poetry_indent
                maintext().insert(
                    line_start,
                    f'<div class="verse indent{indent}">',
                )
                maintext().insert(line_end, replacement + "</div>")
                add_indent_to_css(indent)
                if not in_stanza:  # start of stanza
                    maintext().insert(line_start, '<div class="stanza">')
                    in_stanza = True
            elif in_stanza:  # blank line - end of stanza
                maintext().insert(f"{line_start} -1l lineend", "</div>")
                in_stanza = False
            continue

        # "/l" --> list until we get "l/"
        if selection_lower == "/l":  # open
            maintext().replace(
                line_start,
                line_end,
                "<ul>",
            )
            list_flag = True
            markup_start = step
            reset_ibs_dict()
            continue
        if list_flag:
            if selection_lower == "l/":  # close
                maintext().replace(line_start, line_end, "</ul>")
                list_flag = False
            elif selection:
                do_per_line_markup(selection, line_start, line_end, ibs_dict)
                # Now safe to add list markup
                maintext().insert(line_start, "<li>")
                maintext().insert(line_end, "</li>")
            continue

        # "/#" --> enter new level of blockquote until we get "#/"
        if selection.startswith("/#"):  # open
            blockquote_level += 1
            maintext().replace(
                line_start,
                line_end,
                '<div class="blockquot">',
            )
            markup_start = step
            continue
        if selection == "#/":  # close
            if blockquote_level > 0:
                blockquote_level -= 1
                maintext().replace(
                    line_start,
                    line_end,
                    "</div>",
                )
                if in_para:
                    maintext().insert(f"{line_start}-1l lineend", "</p>")
                    in_para = False
            else:
                raise SyntaxError(f"Line {step}: Unmatched close blockquote (#/)")
            continue

        # "/$" --> nowrap until we get "$/"
        if selection == "/$":  # open
            check_illegal_nesting()
            dollar_nowrap_flag = True
            markup_start = step
            maintext().replace(
                line_start,
                line_end,
                "<p>",
            )
            reset_ibs_dict()
            continue
        if dollar_nowrap_flag and selection == "$/":  # close
            dollar_nowrap_flag = False
            maintext().replace(
                line_start,
                line_end,
                "</p>",
            )
            continue
        # "/*" --> nowrap until we get "*/"
        if selection == "/*":  # open
            check_illegal_nesting()
            asterisk_nowrap_flag = True
            markup_start = step
            maintext().replace(
                line_start,
                line_end,
                "<p>",
            )
            reset_ibs_dict()
            continue
        if asterisk_nowrap_flag and selection == "*/":  # close
            asterisk_nowrap_flag = False
            maintext().replace(
                line_start,
                line_end,
                "</p>",
            )
            continue
        # lines within "/$" or "/*" markup
        if dollar_nowrap_flag or asterisk_nowrap_flag:
            do_per_line_markup(selection, line_start, line_end, ibs_dict)
            # Add 0.5em margin per space character
            if n_spaces > 0:
                maintext().insert(
                    line_start,
                    f'<span style="margin-left: {n_spaces * 0.5}em;">',
                )
                maintext().insert(line_end, "</span>")
            maintext().insert(line_end, "<br>")
            continue

        # "/i" --> index until we get "i/"
        if selection_lower.startswith("/i"):  # open
            check_illegal_nesting()
            index_flag = True
            markup_start = step
            maintext().replace(
                line_start,
                line_end,
                '<ul class="index">',
            )
            reset_ibs_dict()
            index_blank_lines = 2  # Force first entry to be start of section
            continue
        if index_flag:
            if selection_lower == "i/":  # close
                index_flag = False
                maintext().replace(
                    line_start,
                    line_end,
                    "</ul>",
                )
                continue
            if selection == "":
                index_blank_lines += 1
                continue
            if index_blank_lines >= 2:  # Section start
                classname = "ifrst"
            elif index_blank_lines == 1:  # Top-level entry
                classname = "indx"
            else:
                classname = f"isub{int((n_spaces + 1) / 2)}"
            # Convert numbers to page links where the formatting guidelines have been followed:
            # 1. Number must be preceded by a comma (optionally close quote) then one or more spaces
            # 2. Number must be no more than 3 digits (to avoid converting dates, like 1910)
            # 3. Page range may be specified by hyphen/ndash and second number following the first
            # 4. Number or range may be marked up italic or bold, e.g. <b>123</b> or <i>123-124</i>
            linked_selection = re.sub(
                r'(,["\'”’]?) +((<[ib]>)?\b(\d{1,3})([-–]\d{1,3})?\b(</[ib]>)?)',
                rf'\1 <a href="#{PAGE_ID_PREFIX}\4">\2</a>',
                selection,
            )
            maintext().replace(
                line_start, line_end, f'<li class="{classname}">{linked_selection}</li>'
            )
            index_blank_lines = 0
            continue

        # "/c" --> center until we get "c/"
        if selection_lower == "/c":  # open
            check_illegal_nesting()
            center_nowrap_flag = True
            markup_start = step
            maintext().replace(
                line_start,
                line_end,
                '<p class="center">',
            )
            reset_ibs_dict()
            continue
        if center_nowrap_flag:
            if selection_lower == "c/":  # close
                center_nowrap_flag = False
                maintext().replace(
                    line_start,
                    line_end,
                    "</p>",
                )
            else:  # lines within "/c" markup
                do_per_line_markup(selection, line_start, line_end, ibs_dict)
                maintext().insert(line_end, "<br>")
            continue

        # "/r" --> right-align block until we get "r/"
        if selection_lower == "/r":  # open
            check_illegal_nesting()
            right_nowrap_flag = True
            markup_start = step
            maintext().replace(
                line_start,
                line_end,
                '<p class="right">',
            )
            reset_ibs_dict()
            # Begin list of line lengths in order to add right-margin indents later
            right_line_lengths = []
            right_block_line_num = maintext().rowcol(line_start).row
            continue
        if right_nowrap_flag:
            if selection_lower == "r/":  # close
                right_nowrap_flag = False
                maintext().replace(
                    line_start,
                    line_end,
                    "</p>",
                )
                # Add the right-margin indents calculated from line lengths
                # Attempt to preserve shape of right margin (0.5em per character)
                max_len = max(right_line_lengths)
                for line_len in right_line_lengths:
                    right_block_line_num += 1
                    right_pad = max_len - line_len
                    if line_len > 0 and right_pad > 0:
                        maintext().insert(
                            f"{right_block_line_num}.0",
                            f'<span style="margin-left: {right_pad * 0.5}em;">',
                        )
                        maintext().insert(f"{right_block_line_num}.end", "</span>")
                    maintext().insert(f"{right_block_line_num}.end", "<br>")
            else:  # lines within "/r" markup
                # Store line length for later right-margin calculations
                # Convert entities back to single character to get length correct
                # Also sup/sub HTML markup don't count towards length
                len_str = re.sub("&[a-z]+?;", "X", selection)
                len_str = re.sub("</?su[pb]>", "", len_str)
                right_line_lengths.append(len(len_str))
                do_per_line_markup(selection, line_start, line_end, ibs_dict)
            continue

        # In chapter heading - store lines in heading until we get 2 blank lines
        if in_chap_heading:
            if selection:
                chap_line = re.sub(" *<h2.+?>", "", selection)
                chap_heading += (" " if chap_heading else "") + chap_line
                chap_head_blanks = 0
            elif chap_head_blanks == 0:  # First blank line
                # May be two part heading, e.g. "Chapter 1|<blank line>|The Start"
                maintext().insert(line_start, "<br>")
                chap_head_blanks += 1
            else:  # Second blank line = end of heading
                # First blank line will have had "<br>" inserted in elif above.
                # Remove that and add </h2> at end of chapter heading.
                maintext().replace(
                    f"{line_start}-2l lineend", f"{line_start}-1l lineend", "</h2>\n"
                )
                auto_toc += f'<a href="#{chap_id}">{chap_heading}</a><br>\n'
                in_chap_heading = False
            continue

        if selection:
            if not in_para:  # start of paragraph
                maintext().insert(line_start, "<p>")
                in_para = True
        elif in_para:  # blank line - end of paragraph
            maintext().insert(f"{line_start} -1l lineend", "</p>")
            in_para = False
        else:
            # blank line - not in paragraph, so might be chapter heading
            chap_check = maintext().get(f"{line_start}-3l", f"{line_start}+1l lineend")
            if re.match(r"\n\n\n\n[^\n]", chap_check):
                # Look ahead to see what's coming next
                # Don't want to treat h1, centered, poetry, etc as an h2 heading
                # even if it has 4 blank lines before it.
                chap_check = chap_check.lstrip("\n")
                if re.match(
                    r"<h1>|\[Illustration|\[Sidenote|\[Footnote|/[$*#cfilprx]",
                    chap_check,
                    flags=re.IGNORECASE,
                ):
                    continue
                chap_id = make_anchor(
                    maintext().get(f"{line_start}+1l", f"{line_start}+1l lineend")
                )
                maintext().insert(
                    f"{line_start}+1l linestart", f'<h2 class="nobreak" id="{chap_id}">'
                )
                in_chap_heading = True
                chap_start = step
                chap_head_blanks = 0
                chap_heading = ""
    # End of line-at-a-time loop

    # May hit end of file without a final blank line
    if in_para:
        maintext().insert(tk.END, "</p>")

    if pre_flag:
        raise SyntaxError(
            f"Line {markup_start}: Pre-formatted (/x) markup not closed by end of file"
        )
    if front_flag:
        raise SyntaxError(
            f"Line {markup_start}: Frontmatter (/f) markup not closed by end of file"
        )
    if poetry_flag:
        raise SyntaxError(
            f"Line {markup_start}: Poetry (/p) markup not closed by end of file"
        )
    if list_flag:
        raise SyntaxError(
            f"Line {markup_start}: List (/l) markup not closed by end of file"
        )
    if dollar_nowrap_flag:
        raise SyntaxError(
            f"Line {markup_start}: No rewrap (/$) markup not closed by end of file"
        )
    if asterisk_nowrap_flag:
        raise SyntaxError(
            f"Line {markup_start}: No rewrap (/*) markup not closed by end of file"
        )
    if index_flag:
        raise SyntaxError(
            f"Line {markup_start}: Index (/i) markup not closed by end of file"
        )
    if blockquote_level > 0:
        raise SyntaxError(
            f"Line {markup_start}: Blockquote (/#) not closed by end of file"
        )
    if in_chap_heading:
        raise SyntaxError(
            f"Line {chap_start}: Chapter heading does not have two blank lines before end of file"
        )

    # Add autogenerated ToC
    if auto_toc:
        maintext().insert(
            contents_start,
            "\n<!-- Autogenerated TOC. Modify or delete as required. -->\n"
            f"<p>\n{auto_toc}</p>\n"
            "<!-- End Autogenerated TOC. -->\n\n",
        )

    insert_header_footer()
    flush_css_indents()


def do_per_line_markup(
    selection: str,
    line_start: str,
    line_end: str,
    ibs_flags: dict[str, bool],
) -> None:
    """Add <i>, <b>, <sc> markup to this line if needed.

    Some constructions, e.g. poetry, use divs/spans for each line, which
    means that italic, bold or smcap cannot span line breaks, even if
    it was marked up that way at DP. Therefore need to stop/restart these
    inline markups on every line.

    Args:
        selection: Text of the line currently being worked on.
        line_start: Index of start of line.
        line_end: Index of end of line.
        ibs_flags: Global dictionary of flags keeping track of which markups are active.
    """
    if not selection:
        return
    for ch in ibs_flags.keys():
        # Add open at start if already in markup when we got to this line
        if ibs_flags[ch]:
            maintext().insert(line_start, f"<{ch}>")
        # Check status at end of line
        open_idx = selection.rfind(f"<{ch}>")
        close_idx = selection.rfind(f"</{ch}>")
        if open_idx > close_idx:
            ibs_flags[ch] = True  # Markup was left open
        if close_idx > open_idx:
            ibs_flags[ch] = False  # Markup was left closed
        # Add close at end of line if markup was left open
        if ibs_flags[ch]:
            maintext().insert(line_end, f"</{ch}>")


def end_of_css() -> str:
    """Find </style>, marking end of css.

    Returns:
        Index of beginning of line containing </style>
    """
    style_end = maintext().search("</style>", "1.0", regexp=False)
    if not style_end:
        raise SyntaxError("No '</style>' line found in default HTML header")
    return maintext().index(f"{style_end} linestart")


def add_indent_to_css(indent: int) -> None:
    """Add CSS for indent to CSS."""
    css_indents.add(indent)


def flush_css_indents() -> None:
    """Output saved indents to CSS section of file."""
    if not css_indents:
        return

    css_strings: list[str] = ["\n/* Poetry indents */\n"]
    for indent in sorted(css_indents):
        # Default verse with no extra spaces has 3em padding, and -3em text-indent to
        # give hanging indent in case of a continuation line.
        # Every two spaces causes +1em indent, but continuation lines need to align at 3em,
        # so, add the half the space indent to -3em to set the em text-indent for each line
        # For example, if 4 space indent, use 4 * 0.5 - 3 = -1em text-indent, i.e.
        #    .poetry .indent4 {text-indent: -1em;}
        css_strings.append(
            f".poetry .indent{indent} {{text-indent: {indent * 0.5 - 3}em;}}\n"
        )
    css_strings.append("\n")
    maintext().insert(end_of_css(), "".join(css_strings))


def poetry_indentation(poem: str) -> int:
    """Return how much whole poem is already indented by."""
    min_indent = 1000
    for line in poem.splitlines():
        if strip_line := line.lstrip():
            min_indent = min(min_indent, len(line) - len(strip_line))
            if min_indent == 0:
                break
    return min_indent


def html_convert_sub_super(selection: str, line_start: str, line_end: str) -> str:
    """Convert all sub/superscripts in given line & return
    the modified line for later processing.

    Supports `_{subscript}`, `^{superscript}`, and `^x` (single char superscript).

    Args:
        selection: The text of one line.
        line_start: Index of start of line.
        line_end: Index of end of line.

    Returns:
        The modified text of the line.
    """
    selection, nsubs = re.subn(r"_\{(.+?)\}", r"<sub>\1</sub>", selection)
    selection, nsups1 = re.subn(r"\^{(.+?)\}", r"<sup>\1</sup>", selection)
    selection, nsups2 = re.subn(r"\^(.)", r"<sup>\1</sup>", selection)
    if nsubs or nsups1 or nsups2:
        maintext().replace(line_start, line_end, selection)
    return selection


def html_convert_tb(selection: str, line_start: str, line_end: str) -> int:
    """Convert asterisk or <tb> thoughtbreak markup to <hr>.

    Args:
        selection: The text of one line.
        line_start: Index of start of line.
        line_end: Index of end of line.

    Returns:
        True if line was a thoughtbreak.
    """
    selection, nsubs = re.subn(r"^(       \*){5}$", '<hr class="tb">', selection)
    if nsubs:
        maintext().replace(line_start, line_end, selection)
    else:
        selection, nsubs = re.subn("<tb>", '<hr class="tb">', selection)
        if nsubs:
            maintext().replace(line_start, line_end, selection)
    return nsubs


def insert_header_footer() -> None:
    """Insert the default and/or user HTML header, and the footer.

    User's header can either be a complete header, or just some CSS
    to be inserted at the end of the default CSS.
    """
    # Temporarily set the gravity on page markers at the start
    # of the file to RIGHT to ensure the header gets added above them.
    # Need to create list of marks first before changing any gravities,
    # because if pages 1 & 2 are both at the start, then changing the
    # gravity of page 1 to RIGHT means that page 2 is no longer the
    # "next" mark after page 1 (because page 1 is to the right of page 2).
    # Then page 2 would be skipped and remain with LEFT gravity.
    page_mark_list = []
    page_mark = "1.0"
    while page_mark := maintext().page_mark_next(page_mark):
        if maintext().compare(page_mark, ">", "1.0"):
            break
        page_mark_list.append(page_mark)
    for page_mark in page_mark_list:
        maintext().mark_gravity(page_mark, tk.RIGHT)

    # Get user's header file if there is one
    user_path = Path(preferences.prefsdir, HTML_HEADER_NAME)
    if user_path.is_file():
        user_header = replace_header_keywords(user_path.read_text(encoding="utf-8"))
    else:
        user_header = ""
    # If user has provided complete header, insert at start instead of default
    if user_header.startswith("<!DOCTYPE"):
        maintext().insert("1.0", f"{user_header}\n")
    else:
        # Insert default header at start
        default_path = DEFAULT_HTML_DIR.joinpath(HTML_HEADER_NAME)
        default_header = replace_header_keywords(
            default_path.read_text(encoding="utf-8")
        )
        maintext().insert("1.0", f"{default_header}\n")
        # Insert user header if there is one, just before closing "</style>"
        if user_header:
            maintext().insert(end_of_css(), f"{user_header}\n")
    # Insert footer
    maintext().insert(tk.END, "\n</body>\n</html>\n")
    # Restore mark gravity if changed above
    for page_mark in page_mark_list:
        maintext().mark_gravity(page_mark, tk.LEFT)


def replace_header_keywords(header: str) -> str:
    """Replace TITLE and BOOKLANG in the given header to
    their values.

    Args:
        header: Header potentially containing keywords.

    Returns:
        Header with keywords replaced.
    """
    assert HTMLGeneratorDialog.book_title is not None
    header = header.replace("TITLE", HTMLGeneratorDialog.book_title.get())
    header = header.replace("BOOKLANG", maintext().get_language_list()[0])
    return header


def make_anchor(string: str) -> str:
    """Make a valid HTML id from string which may contain HTML markup,
    accented characters, punctuation, etc.

    Args:
        string - String to be converted to HTML id.

    Returns:
        Valid HTML id.
    """
    # Remove any accents
    string = DiacriticRemover.remove_diacritics(string)
    # Clean up dashes and HTML entities
    string = re.sub(r"\p{Dash_Punctuation}+", "-", string)
    string = re.sub("&[a-z]+;", "_", string)
    # Remove super/subscripted text as well as other HTML markup
    string = re.sub("<su[bp]>.+?</subp>", "", string)
    string = re.sub(r"</?[\p{Letter}\p{Number}]+?>", "", string)
    # Remove anything else undesirable
    string = re.sub(r"[^-_\p{Letter}\p{Number}\p{Separator}]", "", string)
    # Replace spaces, etc. with _
    string = re.sub(r"\p{Separator}+", "_", string)
    # Replace multiple underscores with single
    string = re.sub(r"__+", "_", string)
    return string


def get_title() -> str:
    """Determine the book title from the text, if possible.

    Returns:
        Best guess at book title.
    """
    complete_title = ""
    in_title = False
    # Only search first 500 lines of book
    for step in range(1, 500):
        if maintext().compare(f"{step}.0", ">=", tk.END):
            break
        selection = maintext().get(f"{step}.0", f"{step}.end")
        # Check if title complete (blank line or end of frontmatter markup)
        if in_title and (
            not selection or re.match("f/", selection, flags=re.IGNORECASE)
        ):
            break
        # Skip blank lines, illos or block markup
        if not selection or re.match(H1_SKIP_REGEX, selection, flags=re.IGNORECASE):
            continue
        # Strip inline markup, compress multiple spaces & trim trailing spaces
        selection = re.sub(r"<.+?>", "", selection)
        selection = re.sub(r"  +", " ", selection).strip()
        complete_title = f"{complete_title}{selection} "
        in_title = True

    # If no lowercase letters, title is ALL CAPS, so make all but first char lowercase
    # (may not be correct, but best we can do).
    if not re.search(r"\p{Lowercase_Letter}", complete_title):
        complete_title = complete_title.capitalize()
    # Don't want trailing space, comma or period
    return complete_title.rstrip(" ,.")


def adjust_pagemark_postions() -> None:
    """Move page marks to start of next line if at end of non-empty line.

    In particular, we don't want want page mark at the end of a paragraph,
    and hence getting trapped before closing `</p>` markup.
    """

    def mark_list_at_index(mark: str, index: str) -> list[str]:
        """Return a list of the page marks after given mark that are at the given index."""
        mark_list: list[str] = []
        while mark := maintext().page_mark_next(mark):
            if maintext().compare(mark, ">", index):
                break
            mark_list.append(mark)
        return mark_list

    mark = "1.0"
    while mark := maintext().page_mark_next(mark):
        eol = maintext().compare(mark, "==", f"{mark} lineend")
        nonempty = maintext().get(f"{mark} linestart", f"{mark} lineend")
        if eol and nonempty:
            move_mark_list = mark_list_at_index(mark, mark)
            move_mark_list.insert(0, mark)
            next_linestart = maintext().index(f"{mark}+1l linestart")
            # Need to change gravity of this mark and all marks at
            # start of next line, to preserve mark order
            next_mark_list = mark_list_at_index(mark, next_linestart)
            for next_line_mark in next_mark_list:
                maintext().mark_gravity(next_line_mark, tk.RIGHT)
            for move_mark in reversed(move_mark_list):
                maintext().mark_set(move_mark, next_linestart)
            # Restore gravities
            for next_line_mark in next_mark_list:
                maintext().mark_gravity(next_line_mark, tk.LEFT)
            mark = move_mark_list[-1]


def html_convert_page_anchors() -> None:
    """Add page anchors, comments, etc., at page marks depending on dialog settings."""
    page_details = the_file().page_details
    the_file().update_page_marks(page_details)

    # We want page labels sorted by png even if their order in the file is wrong
    # Work in reverse so inserted text doesn't affect later indexes
    page_detail_buffer: list[PageDetail] = []
    last_mark_index = maintext().index("end")

    def lbl_to_pgnum(label: str) -> str:
        """Convert page label to page number by removing prefix."""
        return label.removeprefix(PAGE_LABEL_PREFIX)

    def safe_index(pnum_index: str) -> str:
        """Given an index where a pagenum span is to be inserted,
        move it if necessary, so it won't occur mid-word.

        Args:
            pnum_index: Index where pagenum may be inserted.

        Returns:
            Safe index, which is after pnum_index by as far is necessary
            to avoid mid-word page spans.
        """
        pnum_rowcol = IndexRowCol(pnum_index)
        if pnum_rowcol.col > 0:  # Not at beginning of line
            # Check from previous char to end of line - may be a space before the index point
            check_str = maintext().get(f"{pnum_index} -1c", f"{pnum_index} lineend")
            if len(check_str) > 1:  # Not at end of line
                # Truncate from first space or HTML markup opening
                move = len(re.sub(r"[\s<].*", "", check_str)) - 1
                if move > 0:
                    pnum_index = maintext().index(f"{pnum_index}+{move}c")
        return pnum_index

    def outside_paragraph(index: str) -> bool:
        """Return True if given index is not within `<p>` markup."""
        prev_markup = maintext().find_match(
            r"(<p[> ]|</p>)", IndexRange(index, "1.0"), regexp=True, backwards=True
        )
        return prev_markup is None or maintext().get_match_text(prev_markup) == "</p>"

    def flush_page_detail_buffer() -> None:
        """Output contents of buffer in pgnum span."""
        pgnum = lbl_to_pgnum(page_detail_buffer[0]["label"])
        pstring = f"[{PAGE_LABEL_PREFIX}{pgnum}]" if show_page_numbers else ""
        if len(page_detail_buffer) == 1:
            pagenum_span = (
                f'<span class="pagenum" id="{PAGE_ID_PREFIX}{pgnum}">{pstring}</span>'
            )
        else:
            anchors = f'"></a><a id="{PAGE_ID_PREFIX}'.join(
                [lbl_to_pgnum(pd["label"]) for pd in reversed(page_detail_buffer)]
            )
            anchors = f'<a id="{PAGE_ID_PREFIX}{anchors}"></a>'
            pagenum_span = f'<span class="pagenum">{anchors}{pstring}</span>'
        insert_index = safe_index(page_detail_buffer[0]["index"])
        if outside_paragraph(insert_index):
            pagenum_span = f"<p>{pagenum_span}</p>"
        maintext().insert(insert_index, pagenum_span)

    show_page_numbers = preferences.get(PrefKey.HTML_SHOW_PAGE_NUMBERS)
    for _, page_detail in sorted(page_details.items(), reverse=True):
        label = lbl_to_pgnum(page_detail["label"])
        if not label:
            continue
        # Check if this page mark is effectively at different place to last,
        # i.e. if there are non-space characters between them.
        # If it's a new location for page marks - flush the buffer, writing
        # one pagenum span, with an anchor for each page break if multiple coincident breaks.
        text_between = maintext().get(page_detail["index"], last_mark_index)
        last_mark_index = page_detail["index"]
        if page_detail_buffer and re.search(r"\S", text_between):
            flush_page_detail_buffer()
            page_detail_buffer = []
        page_detail_buffer.append(page_detail)
    if page_detail_buffer:
        flush_page_detail_buffer()


def html_convert_sidenotes() -> None:
    """Convert sidenote markup to HTML divs.

    If multi-paragraph, retain the `<p>` markup, otherwise discard.
    """
    sidenote_end = "1.0"
    while sidenote_start := maintext().search("<p>[Sidenote: ", sidenote_end, tk.END):
        sidenote_end = maintext().search("]</p>", sidenote_start, tk.END)
        if not sidenote_end:
            raise SyntaxError(
                f"Unclosed sidenote:\n{maintext().get(f'{sidenote_start}+3c', f'{sidenote_start} lineend')}"
            )
        sidenote_text = maintext().get(sidenote_start, sidenote_end)
        p_open = p_close = ""
        if "\n\n" in sidenote_text:
            p_open = "<p>"
            p_close = "</p>"
        maintext().replace(
            sidenote_start, f"{sidenote_start}+14c", f'<div class="sidenote">{p_open}'
        )
        maintext().replace(sidenote_end, f"{sidenote_end}+5c", f"{p_close}</div>")


def html_convert_footnotes() -> None:
    """Convert `[Footnote` markup to HTML.

    For each footnote, search backward to find another identically numbered one.
    Then between those two, search for suitable anchors. Link from all anchors
    to the footnote, and from the footnote to the first anchor.
    """
    fn_end = "1.0"
    fn_number = 0
    # Find start of each footnote
    while fn_start := maintext().search(
        r"<p>\[Footnote .+?:", fn_end, tk.END, regexp=True
    ):
        # Find end of footnote
        fn_end = maintext().search(r"]</p>$", fn_start, tk.END, regexp=True)
        if not fn_end:
            raise SyntaxError(
                f"No paragraph break at footnote end:\n{maintext().get(f'{fn_start}+3c', f'{fn_start} lineend')}"
            )
        # Extract label for footnote
        fn_label_end = maintext().search(":", fn_start, tk.END)
        fn_label = maintext().get(f"{fn_start}+13c", fn_label_end)
        fn_number += 1
        fn_id = f"Footnote_{fn_label}_{fn_number}"
        an_id = f"FNanchor_{fn_label}_{fn_number}"
        # Add markup for footnote
        maintext().replace(fn_end, f"{fn_end} lineend", "</p></div>")
        maintext().replace(
            fn_start,
            f"{fn_label_end}+1c",
            f'<div class="footnote"><p><a id="{fn_id}" href="#{an_id}" class="label">[{fn_label}]</a>',
        )

        # Search backwards for another footnote with the same label (or start of file)
        an_search_start = maintext().search(
            f'class="label">[{fn_label}]</a>', fn_start, "1.0", backwards=True
        )
        if an_search_start:
            an_search_start += (
                "+1l"  # Don't want to mistake label of this footnote for an anchor
            )
        else:
            an_search_start = "1.0"
        # For all matching anchors in that range, add HTML markup to anchor, with id on the first only
        id_markup = f'id="{an_id}" '
        while an_start := maintext().search(f"[{fn_label}]", an_search_start, fn_start):
            an_end = f"{an_start}+{len(fn_label) + 2}c"
            maintext().insert(an_end, "</a>")
            open_markup = f'<a {id_markup}href="#{fn_id}" class="fnanchor">'
            maintext().insert(an_start, open_markup)
            id_markup = ""
            an_search_start = maintext().index(f"{an_start}+{len(open_markup) + 4}c")


def html_convert_footnote_landing_zones() -> None:
    """Add HTML markup around Landing Zones beginning with "FOOTNOTES"."""
    lz_end = "1.0"
    while lz_start := maintext().search("<p>FOOTNOTES:</p>", lz_end, tk.END):
        # Find next LZ (or end of file) to help find the end of this LZ
        lz_next = maintext().search("<p>FOOTNOTES:</p>", f"{lz_start}+17c", tk.END)
        if not lz_next:
            lz_next = tk.END
        # Find last footnote in this LZ by searching backwards from next LZ
        lz_end = maintext().search(
            '<div class="footnote">', lz_next, lz_start, backwards=True
        )
        if lz_end:
            lz_end = maintext().search(
                "</div>", lz_end, lz_next
            )  # End of last footnote
        else:
            # No FN in LZ
            lz_end = f"{lz_start}"
        lz_end += " lineend"
        maintext().insert(lz_end, "</div>")
        # <p>FOOTNOTES:</p> ==> <div class="footnotes"><h3>FOOTNOTES:</h3>
        maintext().replace(f"{lz_start}+15c", f"{lz_start}+16c", "h3")
        maintext().replace(lz_start, f"{lz_start}+3c", '<div class="footnotes"><h3>')


def html_add_chapter_divs() -> None:
    """Add chapter divs where there are h2 headings, attempting to
    enclose pagenum spans where appropriate."""
    h2_end = "1.0"
    while h2_start := maintext().search("<h2", h2_end, tk.END):
        h2_end = f'{maintext().search("</h2>", h2_start, tk.END)} lineend'
        prev_page_break = maintext().search(
            '<span class="pagenum"', h2_start, "1.0", backwards=True
        )
        if prev_page_break:
            check_text = maintext().get(f"{prev_page_break} linestart", h2_start)
            check_text = re.sub(r'<span class="pagenum".+?</span>', "", check_text)
            check_text = re.sub(r"</?p>", "", check_text)
            check_text = re.sub(r"\[Illustration.*?\]", "", check_text)
            check_text = check_text.replace("\n", "")
            if not check_text:
                h2_start = f"{prev_page_break} linestart"
        maintext().insert(
            h2_start, '<hr class="chap x-ebookmaker-drop"><div class="chapter">'
        )
        maintext().insert(h2_end, "</div>")


def html_wrap_long_lines() -> None:
    """Wrap lines that are longer than 200 characters, typically where an index line
    has many linked page numbers."""
    wrap_limit = 200
    index = "1.0"
    # Find lines longer than 200 characters
    while index := maintext().search(
        rf"^[^\n]{{{wrap_limit},}}", f"{index} lineend", tk.END, regexp=True
    ):
        # Find previous close tag with a space after (space is possibly preceded by other text, but not an open tag)
        count = tk.IntVar()
        if insert_point := maintext().search(
            "</[a-zA-Z0-9]+?>[^ <]* ",
            f"{index}+{wrap_limit}c",
            index,
            regexp=True,
            backwards=True,
            count=count,
        ):
            # Insert a newline and three extra spaces before the space that was found
            maintext().insert(f"{insert_point}+{count.get() - 1}c", "\n   ")



"""Spell checking functionality"""

import importlib.resources
import logging
from pathlib import Path
from tkinter import ttk
from typing import Callable, Optional, Any
import regex as re

from guiguts.data import dictionaries
from guiguts.file import ProjectDict
from guiguts.checkers import CheckerDialog, CheckerEntry
from guiguts.maintext import maintext, FindMatch
from guiguts.misc_tools import tool_save
from guiguts.preferences import preferences, PersistentInt, PrefKey
from guiguts.utilities import (
    IndexRowCol,
    IndexRange,
    load_wordfile_into_dict,
    cmd_ctrl_string,
    process_accel,
)
from guiguts.widgets import ToolTip

logger = logging.getLogger(__package__)

DEFAULT_DICTIONARY_DIR = importlib.resources.files(dictionaries)

SPELL_CHECK_OK_YES = 0
SPELL_CHECK_OK_NO = 1
SPELL_CHECK_OK_BAD = 2

_the_spell_checker: Optional["SpellChecker"] = None  # pylint: disable=invalid-name


class DictionaryNotFoundError(Exception):
    """Raised when no dictionary found to match languages."""

    def __init__(self, language: str) -> None:
        self.language = language


class SpellingError(FindMatch):
    """Class containing info about spelling error.

    Attributes:
        rowcol: Location of start of word in file.
        count: Number of characters in word.
        word: Word spelt wrongly.
        frequency: Number of occurrences of word.
        bad_word: True if word is in project's bad word list
    """

    def __init__(self, index: IndexRowCol, word: str, bad_word: bool):
        """Initialize SpellingError class."""
        super().__init__(index, len(word))
        self.word = word
        self.frequency = 0
        self.bad_word = bad_word


class SpellCheckerDialog(CheckerDialog):
    """Spell Checker dialog."""

    manual_page = "Tools_Menu#Spelling"

    def __init__(self, **kwargs: Any) -> None:
        """Initialize Spell Checker dialog."""
        assert "add_global_word_callback" in kwargs
        self.add_global_word_callback = kwargs["add_global_word_callback"]
        del kwargs["add_global_word_callback"]
        super().__init__(
            "Spelling Check Results",
            tooltip="\n".join(
                [
                    "Left click: Select & find spelling error",
                    "Right click: Remove spelling error from list",
                    "Shift Right click: Remove all occurrences of spelling error from list",
                    f"With {cmd_ctrl_string()} key: Also add spelling to project dictionary",
                ]
            ),
            **kwargs,
        )
        frame = ttk.Frame(self.header_frame)
        frame.grid(column=0, row=1, sticky="NSEW", pady=5)
        ttk.Label(
            frame,
            text="Threshold ≤ ",
        ).grid(column=0, row=0, sticky="NSW")
        threshold_spinbox = ttk.Spinbox(
            frame,
            textvariable=PersistentInt(PrefKey.SPELL_THRESHOLD),
            from_=1,
            to=999,
            width=4,
        )
        threshold_spinbox.grid(column=1, row=0, sticky="NW", padx=(0, 10))
        ToolTip(
            threshold_spinbox,
            "Do not show errors that appear more than this number of times",
        )

        def invoke_and_break(button: ttk.Button) -> str:
            """Invoke a button and return "break" to avoid further callbacks."""
            button.invoke()
            return "break"

        def add_to_global_dict() -> None:
            """Add current word to global dictionary."""
            current_index = self.current_entry_index()
            if current_index is None:
                return
            checker_entry = self.entries[current_index]
            if checker_entry.text_range:
                word = checker_entry.text.split(maxsplit=1)[0]
                self.add_global_word_callback(word)
                checker = get_spell_checker()
                assert checker is not None
                checker.dictionary[word] = True
            self.remove_entry_current(all_matching=True)

        lang = maintext().get_language_list()[0]
        global_dict_button = ttk.Button(
            frame,
            text=f"Add to Global Dict ({lang})",
            command=add_to_global_dict,
        )
        global_dict_button.grid(column=2, row=0, sticky="NSW")
        for accel in ("Cmd/Ctrl+a", "Cmd/Ctrl+A"):
            _, key_event = process_accel(accel)
            self.bind(key_event, lambda _: invoke_and_break(global_dict_button))
        ToolTip(
            global_dict_button,
            f"{cmd_ctrl_string()}+A",
            use_pointer_pos=True,
        )
        project_dict_button = ttk.Button(
            frame,
            text="Add to Project Dict",
            command=lambda: self.process_remove_entry_current(all_matching=True),
        )
        project_dict_button.grid(column=3, row=0, sticky="NSW")
        for accel in ("Cmd/Ctrl+p", "Cmd/Ctrl+P"):
            _, key_event = process_accel(accel)
            self.bind(key_event, lambda _: invoke_and_break(project_dict_button))
        ToolTip(
            project_dict_button,
            f"{cmd_ctrl_string()}+P",
            use_pointer_pos=True,
        )
        skip_button = ttk.Button(
            frame,
            text="Skip",
            command=lambda: self.remove_entry_current(all_matching=False),
        )
        skip_button.grid(column=4, row=0, sticky="NSW")
        for accel in ("Cmd/Ctrl+s", "Cmd/Ctrl+S"):
            _, key_event = process_accel(accel)
            self.bind(key_event, lambda _: invoke_and_break(skip_button))
        ToolTip(
            skip_button,
            f"{cmd_ctrl_string()}+S",
            use_pointer_pos=True,
        )
        skip_all_button = ttk.Button(
            frame,
            text="Skip All",
            command=lambda: self.remove_entry_current(all_matching=True),
        )
        skip_all_button.grid(column=5, row=0, sticky="NSW")
        for accel in ("Cmd/Ctrl+i", "Cmd/Ctrl+I"):
            _, key_event = process_accel(accel)
            self.bind(key_event, lambda _: invoke_and_break(skip_all_button))
        ToolTip(
            skip_all_button,
            f"{cmd_ctrl_string()}+I",
            use_pointer_pos=True,
        )


class SpellChecker:
    """Provides spell check functionality."""

    def __init__(self) -> None:
        """Initialize SpellChecker class."""
        self.dictionary: dict[str, bool] = {}
        self.language_list = maintext().get_language_list()
        for lang in self.language_list:
            self.add_words_from_language(lang)

    def do_spell_check(self, project_dict: ProjectDict) -> list[SpellingError]:
        """Spell check the currently loaded file, or just the selected range(s).

        Returns:
            List of spelling errors.
        """
        spelling_errors = []
        spelling_counts: dict[str, int] = {}

        minrow = mincol = maxrow = maxcol = 0
        if sel_ranges := maintext().selected_ranges():
            minrow = sel_ranges[0].start.row
            mincol = sel_ranges[0].start.col
            maxrow = sel_ranges[-1].end.row
            maxcol = sel_ranges[-1].end.col
        column_selection = len(sel_ranges) > 1
        for line, line_num in maintext().get_lines():
            # Handle doing selection only
            if sel_ranges:
                # If haven't reached the line range, skip
                if line_num < minrow:
                    continue
                # If past the line range, stop
                if line_num > maxrow:
                    break
                # Clear the columns outside the selection
                if column_selection or line_num == minrow:
                    line = mincol * " " + line[mincol:]
                if column_selection or line_num == maxrow:
                    line = line[:maxcol]

            words = re.split(r"[^\p{Alnum}\p{Mark}'’]", line)
            next_col = 0
            for word in words:
                col = next_col
                next_col = col + len(word) + 1
                if word:
                    spell_check_result = self.spell_check_word(word, project_dict)

                    # If word has leading straight apostrophe, it might be
                    # open single quote; trim it and check again
                    if spell_check_result == SPELL_CHECK_OK_NO and word.startswith("'"):
                        word = word[1:]
                        col += 1
                        spell_check_result = self.spell_check_word(word, project_dict)

                    # If trailing straight/curly apostrophe, it might be
                    # close single quote; trim it and check again
                    if spell_check_result == SPELL_CHECK_OK_NO and re.search(
                        r"['’]$", word
                    ):
                        word = word[:-1]
                        spell_check_result = self.spell_check_word(word, project_dict)

                    # If not found in dictionary, and word is not now empty
                    # or only consisting of single quotes, add it to errors list
                    if spell_check_result != SPELL_CHECK_OK_YES and not re.fullmatch(
                        r"['’]*", word
                    ):
                        spelling_errors.append(
                            SpellingError(
                                IndexRowCol(line_num, col),
                                word,
                                spell_check_result == SPELL_CHECK_OK_BAD,
                            )
                        )
                        try:
                            spelling_counts[word] += 1
                        except KeyError:
                            spelling_counts[word] = 1
        # Now all spelling errors have been found, update the frequency of each one
        for spelling in spelling_errors:
            spelling.frequency = spelling_counts[spelling.word]
        return spelling_errors

    def spell_check_word(self, word: str, project_dict: ProjectDict) -> int:
        """Spell check the given word.

        Allows things like numbers, years, decades, etc.

        Args:
            word: Word to be checked.

        Returns:
            SPELL_CHECK_OK_BAD if word is in the bad_words list
            SPELL_CHECK_OK_YES if word is in dictionary
            SPELL_CHECK_OK_NO if neither (wrong spelling)
        """
        word_status = self.spell_check_word_apos(word, project_dict)
        # If it's a bad word or in the dictionary, we're done
        if word_status in (SPELL_CHECK_OK_BAD, SPELL_CHECK_OK_YES):
            return word_status

        # Some languages use l', quest', etc., before word.
        # Accept if the "prefix" and the remainder are both good.
        # Prefix can be with or without apostrophe, but dictionary
        # is safer if prefix has apostrophe to avoid prefix being
        # accepted it it is used as a standalone word.
        match = re.match(r"(?P<prefix>\w+)['’](?P<remainder>\w+)", word)
        if (
            match is not None
            and self.spell_check_word_apos(match.group("remainder"), project_dict)
            == SPELL_CHECK_OK_YES
            and (
                self.spell_check_word_apos(match.group("prefix"), project_dict)
                == SPELL_CHECK_OK_YES
                or self.spell_check_word_apos(match.group("prefix") + "'", project_dict)
                == SPELL_CHECK_OK_YES
                or self.spell_check_word_apos(match.group("prefix") + "’", project_dict)
                == SPELL_CHECK_OK_YES
            )
        ):
            return SPELL_CHECK_OK_YES

        # Now check numbers
        if (
            # word is all digits
            re.fullmatch(r"\d+", word)  # pylint: disable=too-many-boolean-expressions
            # ...1st, ...21st, ...31st, etc
            or re.fullmatch(r"(\d*[02-9])?1st", word, flags=re.IGNORECASE)
            # ...2nd, ...22nd, ...32nd, etc (also 2d, 22d, etc)
            or re.fullmatch(r"(\d*[02-9])?2n?d", word, flags=re.IGNORECASE)
            # ...3rd, ...23rd, ...33rd, etc (also 3d, 33d, etc)
            or re.fullmatch(r"(\d*[02-9])?3r?d", word, flags=re.IGNORECASE)
            # ...0th, ...4th, ...5th, etc
            or re.fullmatch(r"\d*[04-9]th", word, flags=re.IGNORECASE)
            # ...11th, ...12th, ...13th
            or re.fullmatch(r"\d*1[123]th", word, flags=re.IGNORECASE)
        ):
            return SPELL_CHECK_OK_YES

        # Allow decades/years
        if (
            # e.g. '20s or 20s (abbreviation for 1820s)
            re.fullmatch(r"['’]?\d\ds", word)
            # e.g. '62 (abbreviation for 1862)
            or re.fullmatch(r"['’]\d\d", word)
            # e.g. 1820s
            or re.fullmatch(r"1\d{3}s", word)
        ):
            return SPELL_CHECK_OK_YES

        # Allow abbreviations for shillings and pence (not pounds
        # because 20l is common scanno for the number 201)
        # e.g. 15s or 6d (up to 2 digits of old English shillings and pence)
        if re.fullmatch(r"\d{1,2}[sd]", word):
            return SPELL_CHECK_OK_YES

        # <sc> DP markup
        if re.fullmatch(r"sc", word, flags=re.IGNORECASE):
            return SPELL_CHECK_OK_YES

        return SPELL_CHECK_OK_NO

    def spell_check_word_apos(self, word: str, project_dict: ProjectDict) -> int:
        """Spell check the given word.

        Copes with straight or curly apostrophes.

        Args:
            word: Word to be checked.

        Returns:
            SPELL_CHECK_OK_BAD if word is in the bad_words list
            SPELL_CHECK_OK_YES if word is in dictionary
            SPELL_CHECK_OK_NO if neither (wrong spelling)
        """
        word_status = self.spell_check_word_case(word, project_dict)
        # If it's a bad word or in the dictionary, we're done
        if word_status in (SPELL_CHECK_OK_BAD, SPELL_CHECK_OK_YES):
            return word_status

        # Otherwise, try swapping apostrophes if there are any, and re-check
        if "'" in word:
            word = re.sub(r"'", r"’", word)
        elif "’" in word:
            word = re.sub(r"’", r"'", word)
        else:
            return word_status
        return self.spell_check_word_case(word, project_dict)

    def spell_check_word_case(self, word: str, project_dict: ProjectDict) -> int:
        """Spell check word, allowing for differences of case.

        Check same case, lower case, or title case (e.g. LONDON matches London)
        Can't just do case-insensitive check because we don't want "london" to be OK.

        Args:
            word: Word to be checked.

        Returns:
            SPELL_CHECK_OK_BAD if word is in the bad_words list
            SPELL_CHECK_OK_YES if word is in dictionary
            SPELL_CHECK_OK_NO if neither (wrong spelling)
        """
        # Check if word is a bad word - case must match for bad words
        if word in project_dict.bad_words:
            return SPELL_CHECK_OK_BAD

        lower_word = word.lower()
        if lower_word == word:
            lower_word = ""
        # Caution converting to title case - we want "LONDON" to match "London",
        # but we don't want "london" to match London, so leave first letter case
        # as-is, and lowercase the rest
        if len(word) > 1:
            title_word = word[0:1] + word[1:].lower()
            if title_word == word:
                title_word = ""
        else:
            title_word = ""

        # Now check global & project dictionaries for good words
        for dictionary in (self.dictionary, project_dict.good_words):
            if (
                word in dictionary
                or (lower_word and lower_word in dictionary)
                or (title_word and title_word in dictionary)
            ):
                return SPELL_CHECK_OK_YES

        return SPELL_CHECK_OK_NO

    def add_words_from_language(self, lang: str) -> None:
        """Add dictionary words for given language.

        Attempts to load default and user dictionaries.
        Outputs warning if default dictionary doesn't exist.
        Raises DictionaryNotFoundError exception if neither dictionary exists.

        Args:
            lang: Language to be loaded.
        """
        path = DEFAULT_DICTIONARY_DIR.joinpath(f"dict_{lang}_default.txt")
        words_loaded = load_wordfile_into_dict(path, self.dictionary)
        if not words_loaded:
            logger.warning(f"No default dictionary for language {lang}")

        path = Path(preferences.prefsdir, f"dict_{lang}_user.txt")
        if load_wordfile_into_dict(path, self.dictionary):
            words_loaded = True
        if not words_loaded:  # Neither default nor user dictionary exist
            raise DictionaryNotFoundError(lang)


def get_spell_checker() -> SpellChecker | None:
    """Avoid duplicate spell checker by returning a SpellChecker object to
    calling tool/application.

    Returns:
        A SpellChecker object if required dictionary present, otherwise None
    """

    global _the_spell_checker

    # If we already have a spell checker with the wrong languages, delete it
    if (
        _the_spell_checker is not None
        and _the_spell_checker.language_list != maintext().get_language_list()
    ):
        _the_spell_checker = None
    if _the_spell_checker is None:
        try:
            _the_spell_checker = SpellChecker()
        except DictionaryNotFoundError as exc:
            logger.error(f"Dictionary not found for language: {exc.language}")
            return None
    return _the_spell_checker


def spell_check(
    project_dict: ProjectDict,
    add_project_word_callback: Callable[[str], None],
    add_global_word_callback: Callable[[str], None],
) -> None:
    """Spell check the currently loaded file."""

    if not tool_save():
        return

    checker = get_spell_checker()
    if checker is None:
        return
    bad_spellings = checker.do_spell_check(project_dict)

    def process_spelling(checker_entry: CheckerEntry) -> None:
        """Process the spelling error by adding the word to the project dictionary."""
        if checker_entry.text_range:
            add_project_word_callback(checker_entry.text.split(maxsplit=1)[0])

    checker_dialog = SpellCheckerDialog.show_dialog(
        rerun_command=lambda: spell_check(
            project_dict,
            add_project_word_callback,
            add_global_word_callback,
        ),
        process_command=process_spelling,
        switch_focus_when_clicked=False,
        add_global_word_callback=add_global_word_callback,
    )

    # Construct opening line describing the search
    sel_only = " (selected text only)" if len(maintext().selected_ranges()) > 0 else ""
    checker_dialog.add_header("Start of Spelling Check" + sel_only, "")

    threshold = preferences.get(PrefKey.SPELL_THRESHOLD)
    for spelling in bad_spellings:
        if spelling.frequency > threshold:
            continue
        end_rowcol = IndexRowCol(
            maintext().index(spelling.rowcol.index() + f"+{spelling.count}c")
        )
        bad_str = " ***" if spelling.bad_word else ""
        checker_dialog.add_entry(
            f"{spelling.word} ({spelling.frequency})" + bad_str,
            IndexRange(spelling.rowcol, end_rowcol),
            0,
            spelling.count,
        )
    checker_dialog.add_footer("", "End of Spelling Check" + sel_only)
    checker_dialog.display_entries()



"""Support running of checking tools"""

from enum import Enum, IntEnum, StrEnum, auto
import tkinter as tk
from tkinter import ttk
from typing import Any, Optional, Callable

import regex as re

from guiguts.maintext import maintext
from guiguts.mainwindow import ScrolledReadOnlyText
from guiguts.preferences import PrefKey
from guiguts.root import root
from guiguts.utilities import (
    IndexRowCol,
    IndexRange,
    is_mac,
    sing_plur,
)
from guiguts.widgets import ToplevelDialog, TlDlg, mouse_bind, Busy, ToolTip

MARK_ENTRY_TO_SELECT = "MarkEntryToSelect"
HILITE_TAG_NAME = "chk_hilite"
ERROR_PREFIX_TAG_NAME = "chk_error_prefix"
REFRESH_MESSAGE = "Click this message to refresh after Undo/Redo"


class CheckerEntryType(Enum):
    """Enum class to store Checker Entry types."""

    HEADER = auto()
    CONTENT = auto()
    FOOTER = auto()


class CheckerEntrySeverity(IntEnum):
    """Enum class to store severity of Checker Entry (info/warning/error/etc).

    IntEnums can be used as integers, and auto() allocates incrementing values,
    so code such as `severity > CheckerEntrySeverity.INFO` is valid.

    If additional severities such as WARNING or CRITICAL are added, they must
    be inserted in the list below to maintain an order of increasing severity.
    """

    OTHER = auto()  # Headers, blank lines, etc.
    INFO = auto()  # Messages that do not indicate a problem
    ERROR = auto()  # Messages that indicate a problem


class CheckerEntry:
    """Class to hold one entry in the Checker dialog.

    Current implementation enforces (on creation) and assumes (in use) that
    no section contains Entries of more than one type.
    """

    def __init__(
        self,
        text: str,
        text_range: Optional[IndexRange],
        hilite_start: Optional[int],
        hilite_end: Optional[int],
        section: int,
        initial_pos: int,
        entry_type: CheckerEntryType,
        error_prefix: str,
    ) -> None:
        """Initialize CheckerEntry object.

        Args:
            text: Single line of text to display in checker dialog.
            text_range: Optional start and end of point of interest in main text widget.
            hilite_start: Optional column to begin higlighting entry in dialog.
            hilite_end: Optional column to end higlighting entry in dialog.
            section: Section entry belongs to.
            initial_pos: Position of entry during initial creation.
            entry_type: Type of entry: header, footer, content, etc.
            error_prefix: Prefix string indicating an error - empty if no error.
        """
        self.text = text
        self.text_range = text_range
        self.hilite_start = hilite_start
        self.hilite_end = hilite_end
        self.section = section
        self.initial_pos = initial_pos
        self.entry_type = entry_type
        self.error_prefix = error_prefix
        self.severity = (
            CheckerEntrySeverity.OTHER
            if text_range is None
            else (
                CheckerEntrySeverity.ERROR
                if error_prefix
                else CheckerEntrySeverity.INFO
            )
        )


class CheckerSortType(StrEnum):
    """Enum class to store Checker Dialog sort types."""

    ALPHABETIC = auto()
    ROWCOL = auto()


class CheckerViewOptionsDialog(ToplevelDialog):
    """Dialog to allow user to show/hide checker message types."""

    def __init__(
        self,
        checker_dialog: "CheckerDialog",
    ) -> None:
        """Create View Options dialog.

        Args:
            checker_dialog: The checker dialog that this dialog belongs to.
        """
        super().__init__(
            f"{checker_dialog.title()} - View Options", resize_x=False, resize_y=False
        )
        self.checker_dialog = checker_dialog

        self.flags: list[tk.BooleanVar] = []
        check_frame = ttk.Frame(self.top_frame)
        check_frame.grid(row=0, column=0)
        max_height = 15  # Don't want too many checkbuttons per column
        for row, option_filter in enumerate(self.checker_dialog.view_options_filters):

            check_var = tk.BooleanVar(value=option_filter.on)

            def btn_clicked(row: int = row, var: tk.BooleanVar = check_var) -> None:
                """Called when filter setting is changed."""
                self.checker_dialog.view_options_filters[row].on = var.get()
                self.checker_dialog.display_entries()

            ttk.Checkbutton(
                check_frame,
                text=option_filter.label,
                command=btn_clicked,
                variable=check_var,
            ).grid(row=row % max_height, column=row // max_height, sticky="NSW")
            self.flags.append(check_var)

        btn_frame = ttk.Frame(self.top_frame)
        btn_frame.grid(row=1, column=0, pady=(5, 0))
        ttk.Button(btn_frame, text="Hide All", command=lambda: self.set_all(True)).grid(
            row=0, column=0, padx=5
        )
        ttk.Button(btn_frame, text="See All", command=lambda: self.set_all(False)).grid(
            row=0, column=1, padx=5
        )

    def set_all(self, value: bool) -> None:
        """Set all the checkbuttons to the given value."""
        for flag in self.flags:
            flag.set(value)
        for row, _ in enumerate(self.checker_dialog.view_options_filters):
            self.checker_dialog.view_options_filters[row].on = value
        self.checker_dialog.display_entries()

    def on_destroy(self) -> None:
        if self.checker_dialog.winfo_exists():
            self.checker_dialog.lift()
        super().on_destroy()


class CheckerFilter:
    """Class to store a single filter used for View Options."""

    def __init__(self, label: str, regex: str, on: bool = False) -> None:
        """Initialize checker filter class.

        Args:
            on: Whether filter is turned on.
        """
        self.label = label
        self.regex = re.compile(regex)
        self.on = on

    def matches(self, entry: CheckerEntry) -> bool:
        """Return whether filter matches given entry."""
        raise NotImplementedError()


class CheckerFilterText(CheckerFilter):
    """Filter based on checker entry text."""

    def matches(self, entry: CheckerEntry) -> bool:
        """Return whether filter matches given entry."""
        return bool(self.regex.fullmatch(entry.text))


class CheckerFilterErrorPrefix(CheckerFilter):
    """Filter based on checker entry error prefix."""

    def matches(self, entry: CheckerEntry) -> bool:
        """Return whether filter matches given entry."""
        return bool(self.regex.fullmatch(entry.error_prefix))


class CheckerDialog(ToplevelDialog):
    """Dialog to show results of running a check.

    Attributes:
        text: Text widget to contain results.
        header_frame: Frame at top of widget containing configuration buttons, fields, etc.
                      Row 0 contains common controls, rows 1+ available for specific dialogs.
        count_label: Label showing how many linked entries there are in the dialog
    """

    # Slight complexity needed.
    # 1. Each type of checker dialog needs its own selection_on_clear flag
    # 2. Flag should be persistent across different instances of dialog, since
    #    some dialogs are deleted & recreated
    # Therefore store as a dict class variable keyed on class name
    selection_on_clear: dict[str, Optional[int]] = {}

    def __init__(
        self,
        title: str,
        rerun_command: Callable[[], None],
        tooltip: str,
        process_command: Optional[Callable[[CheckerEntry], None]] = None,
        sort_key_rowcol: Optional[Callable[[CheckerEntry], tuple]] = None,
        sort_key_alpha: Optional[Callable[[CheckerEntry], tuple]] = None,
        show_suspects_only: bool = False,
        clear_on_undo_redo: bool = False,
        view_options_dialog_class: Optional[type[CheckerViewOptionsDialog]] = None,
        view_options_filters: Optional[list[CheckerFilter]] = None,
        switch_focus_when_clicked: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the dialog.

        Args:
            title: Title for dialog.
            rerun_command: Function to call to re-run the check.
            tooltip: Tooltip to show in messages text widget.
            process_command: Function to call to "process" the current error, e.g. swap he/be.
            sort_key_rowcol: Function to sort by row & column.
            sort_key_alpha: Function to sort by alpha/type.
            show_suspects_only: True to show "Suspects only" checkbutton.
            clear_on_undo_redo: True to cause dialog to be cleared if Undo/Redo are used.
            view_options_dialog_class: Class to use for a View Options dialog.
            view_options_filters: List of filters to control which messages are shown.
        """
        super().__init__(title, **kwargs)
        self.top_frame.rowconfigure(0, weight=0)
        self.header_frame = ttk.Frame(self.top_frame, padding=2)
        self.header_frame.grid(row=0, column=0, sticky="NSEW")

        self.header_frame.columnconfigure(0, weight=1)
        options_frame = ttk.Frame(self.header_frame)
        options_frame.grid(row=0, column=0, sticky="NSEW")
        options_frame.columnconfigure(0, weight=0)
        options_frame.columnconfigure(1, weight=1)
        self.count_label = ttk.Label(options_frame, text="No results")
        self.count_label.grid(row=0, column=0, sticky="NSW")

        def copy_errors() -> None:
            """Copy text messages to clipboard."""
            self.clipboard_clear()
            self.clipboard_append(self.text.get("1.0", tk.END))

        copy_button = ttk.Button(
            options_frame, text="Copy Results", command=copy_errors
        )
        copy_button.grid(row=0, column=1, sticky="NS")

        def rerunner() -> None:
            self.selection_on_clear[self.__class__.__name__] = None
            self.rerun_command()

        self.rerun_command = rerun_command
        self.rerun_button = ttk.Button(options_frame, text="Re-run", command=rerunner)
        self.rerun_button.grid(row=0, column=2, sticky="NSE", padx=(10, 0))

        self.suspects_only_btn: Optional[ttk.Checkbutton]
        if show_suspects_only:
            # Can't use a PersistentBoolean directly, since we save this value for each checker dialog
            suspects_only_var = tk.BooleanVar(
                self, self.get_dialog_pref(PrefKey.CHECKERDIALOG_SUSPECTS_ONLY_DICT)
            )

            def suspects_only_changed() -> None:
                self.save_dialog_pref(
                    PrefKey.CHECKERDIALOG_SUSPECTS_ONLY_DICT, suspects_only_var.get()
                )
                self.display_entries()

            self.suspects_only_btn = ttk.Checkbutton(
                options_frame,
                text="Suspects Only",
                variable=suspects_only_var,
                command=suspects_only_changed,
                takefocus=False,
            )
            self.suspects_only_btn.grid(row=1, column=0, sticky="NSW")
        else:
            self.suspects_only_btn = None

        sort_frame = ttk.Frame(options_frame)
        sort_frame.grid(row=1, column=1, sticky="NS", pady=5)
        sort_frame.rowconfigure(0, weight=1)
        ttk.Label(
            sort_frame,
            text="Sort:",
        ).grid(row=0, column=0, sticky="NS", padx=5)

        # Can't use a PersistentString directly, since we save this value for each checker dialog
        sort_type = tk.StringVar(
            self,
            self.get_dialog_pref(PrefKey.CHECKERDIALOG_SORT_TYPE_DICT)
            or CheckerSortType.ROWCOL,
        )

        def sort_type_changed() -> None:
            self.save_dialog_pref(PrefKey.CHECKERDIALOG_SORT_TYPE_DICT, sort_type.get())
            self.display_entries()

        ttk.Radiobutton(
            sort_frame,
            text="Line & Col",
            command=sort_type_changed,
            variable=sort_type,
            value=CheckerSortType.ROWCOL,
            takefocus=False,
        ).grid(row=0, column=1, sticky="NS", padx=2)
        ttk.Radiobutton(
            sort_frame,
            text="Alpha/Type",
            command=sort_type_changed,
            variable=sort_type,
            value=CheckerSortType.ALPHABETIC,
            takefocus=False,
        ).grid(row=0, column=2, sticky="NS", padx=2)

        self.view_options_dialog: Optional[CheckerViewOptionsDialog] = None

        def show_view_options() -> None:
            """Show the view options dialog."""
            assert view_options_dialog_class is not None
            self.view_options_dialog = view_options_dialog_class.show_dialog(
                checker_dialog=self
            )

        if view_options_dialog_class is not None:
            ttk.Button(
                options_frame, text="View Options", command=show_view_options
            ).grid(row=1, column=2, sticky="NSE", pady=5)
        if view_options_filters is None:
            view_options_filters = []
        self.view_options_filters = view_options_filters

        self.top_frame.rowconfigure(1, weight=1)
        self.text = ScrolledReadOnlyText(
            self.top_frame,
            context_menu=False,
            wrap=tk.NONE,
            font=maintext().font,
        )
        self.text.grid(row=1, column=0, sticky="NSEW")
        ToolTip(self.text, tooltip, use_pointer_pos=True)

        # 3 binary choices:
        #     remove/not_remove (just select) - controlled by button 3 or button 1
        #     process/not_process - controlled by Cmd/Ctrl pressed or not
        #     match/not_match - controlled by Shift pressed or not
        # Only 7 of the 8 possibilities needed, since "select not_process match" makes no sense
        mouse_bind(self.text, "1", self.select_entry_by_click)
        mouse_bind(self.text, "3", self.remove_entry_by_click)
        mouse_bind(
            self.text,
            "Shift+3",
            lambda event: self.remove_entry_by_click(event, all_matching=True),
        )
        mouse_bind(self.text, "Cmd/Ctrl+1", self.process_entry_by_click)
        mouse_bind(
            self.text,
            "Shift+Cmd/Ctrl+1",
            lambda event: self.process_entry_by_click(event, all_matching=True),
        )
        mouse_bind(self.text, "Cmd/Ctrl+3", self.process_remove_entry_by_click)
        mouse_bind(
            self.text,
            "Shift+Cmd/Ctrl+3",
            lambda event: self.process_remove_entry_by_click(event, all_matching=True),
        )
        self.bind("<Up>", lambda _e: self.select_entry_by_arrow(-1))
        self.bind("<Down>", lambda _e: self.select_entry_by_arrow(1))

        self.text.bind("<Home>", lambda _e: self.select_entry_by_index(0))
        self.text.bind("<Shift-Home>", lambda _e: self.select_entry_by_index(0))
        self.text.bind(
            "<End>", lambda _e: self.select_entry_by_index(len(self.entries) - 1)
        )
        self.text.bind(
            "<Shift-End>", lambda _e: self.select_entry_by_index(len(self.entries) - 1)
        )
        # Bind same keys as main window uses for top/bottom on Mac.
        # Above bindings work already for Windows
        if is_mac():
            self.text.bind("<Command-Up>", lambda _e: self.select_entry_by_index(0))
            self.text.bind(
                "<Shift-Command-Up>", lambda _e: self.select_entry_by_index(0)
            )
            self.text.bind(
                "<Command-Down>",
                lambda _e: self.select_entry_by_index(len(self.entries) - 1),
            )
            self.text.bind(
                "<Shift-Command-Down>",
                lambda _e: self.select_entry_by_index(len(self.entries) - 1),
            )

        self.process_command = process_command
        self.rowcol_key = sort_key_rowcol or CheckerDialog.sort_key_rowcol
        self.alpha_key = sort_key_alpha or CheckerDialog.sort_key_alpha
        self.text.tag_configure(
            HILITE_TAG_NAME,
            background=maintext()["selectbackground"],
            foreground=maintext()["selectforeground"],
        )
        self.text.tag_configure(ERROR_PREFIX_TAG_NAME, foreground="red")

        def do_clear_on_undo_redo() -> None:
            """If undo/redo operation should trigger user to Re-run tool, clear dialog."""
            self.selection_on_clear[self.__class__.__name__] = (
                self.current_entry_index()
            )
            self.reset()
            self.text.insert(tk.END, REFRESH_MESSAGE)
            self.update_count_label()

        # Ensure this class has an entry in `selection_on_clear`
        # Only if not there already, because we don't want to overwrite it
        # if this is a new instance of a previously existing dialog, with the old
        # one destroyed and this one created due to a refresh/re-run.
        if self.__class__.__name__ not in self.selection_on_clear:
            self.selection_on_clear[self.__class__.__name__] = None
        # If this dialog should be cleared on undo/redo, add callback to maintext
        if clear_on_undo_redo:
            maintext().add_undo_redo_callback(
                self.__class__.__name__, do_clear_on_undo_redo
            )

        if switch_focus_when_clicked is None:
            switch_focus_when_clicked = not is_mac()
        self.switch_focus_when_clicked = switch_focus_when_clicked

        self.count_linked_entries = 0  # Not the same as len(self.entries)
        self.count_suspects = 0
        self.section_count = 0
        self.selected_text = ""
        self.selected_text_range: Optional[IndexRange] = None
        self.reset()

    def __new__(cls, *args: Any, **kwargs: Any) -> "CheckerDialog":
        """Ensure CheckerDialogs are not instantiated directly."""
        if cls is CheckerDialog:
            raise TypeError(f"only children of '{cls.__name__}' may be instantiated")
        return object.__new__(cls)

    @classmethod
    def show_dialog(
        cls: type[TlDlg],
        **kwargs: Any,
    ) -> TlDlg:
        """Show the instance of this dialog class, or create it if it doesn't exist.

        Args:
            title: Dialog title.
            args: Optional args to pass to dialog constructor.
            kwargs: Optional kwargs to pass to dialog constructor.
        """
        Busy.busy()
        return super().show_dialog(**kwargs)

    @classmethod
    def sort_key_rowcol(
        cls,
        entry: CheckerEntry,
    ) -> tuple[int, int, int]:
        """Default sort key function to sort entries by row/col.

        Preserve entries in sections.
        If header/footer, then preserve original order
        If no row/col, place content lines before those with row/col
        If row/col, sort by row, then col.
        """
        if entry.entry_type == CheckerEntryType.HEADER:
            return (entry.section, entry.initial_pos, 0)
        if entry.entry_type == CheckerEntryType.FOOTER:
            return (entry.section, entry.initial_pos, 0)

        # Content without row/col comes before content with, and preserves its original order
        if entry.text_range is None:
            return (entry.section, 0, entry.initial_pos)
        # Content with row/col is ordered by row, then col
        return (
            entry.section,
            entry.text_range.start.row,
            entry.text_range.start.col,
        )

    @classmethod
    def sort_key_alpha(
        cls,
        entry: CheckerEntry,
    ) -> tuple[int, str, str, int, int]:
        """Default sort key function to sort entries by text, putting identical upper
            and lower case versions together.

        If text is highlighted, use that, otherwise use whole line of text
        Preserve entries in sections.
        If header/footer, then preserve original order, putting header first & footer last.
        Place text with no row/col before identical text with row/col.
        """
        # Force header before content, and footer after content (though should be
        # unnecessary since in different sections) preserving original order
        if entry.entry_type == CheckerEntryType.HEADER:
            return (
                entry.section,
                "",
                "",
                0,
                entry.initial_pos,
            )
        if entry.entry_type == CheckerEntryType.FOOTER:
            return (
                entry.section,
                "\ufeff",
                "\ufeff",
                0,
                entry.initial_pos,
            )

        # Get text for comparison
        if entry.hilite_start is None:
            text = entry.text
        else:
            text = entry.text[entry.hilite_start : entry.hilite_end]
        text_low = text.lower()

        # Content without row/col comes before identical content with row/col
        if entry.text_range is None:
            return (entry.section, text_low, text, 0, 0)
        return (
            entry.section,
            text_low,
            text,
            entry.text_range.start.row,
            entry.text_range.start.col,
        )

    def reset(self) -> None:
        """Reset dialog and associated structures & marks."""
        super().reset()
        self.entries: list[CheckerEntry] = []
        self.count_linked_entries = 0
        self.count_suspects = 0
        self.section_count = 0
        self.selected_text = ""
        self.selected_text_range = None
        self.update_count_label(working=True)
        if self.text.winfo_exists():
            self.text.delete("1.0", tk.END)
        if maintext().winfo_exists():
            maintext().clear_marks(self.get_mark_prefix())
            maintext().remove_spotlights()

    def select_entry_after_undo_redo(self) -> None:
        """Select the saved entry, if any, after a re-run following undo/redo."""
        entry_index = self.selection_on_clear[self.__class__.__name__]
        if entry_index is not None:
            self.select_entry_by_index(entry_index)

    def on_destroy(self) -> None:
        """Override method that tidies up when the dialog is destroyed.

        Needs to remove the undo_redo callback if there is one.
        Also the View Options dialog if it is popped.
        """
        super().on_destroy()
        maintext().remove_undo_redo_callback(self.__class__.__name__)
        if (
            self.view_options_dialog is not None
            and self.view_options_dialog.winfo_exists()
        ):
            self.view_options_dialog.destroy()
            self.view_options_dialog = None

    def new_section(self) -> None:
        """Start a new section in the dialog.

        When entries are sorted, entries within a section remain in the section.
        It's not a problem if a section is empty.
        """
        self.section_count += 1

    def _add_headfoot(self, hf_type: CheckerEntryType, *headfoot_lines: str) -> None:
        """Internal method to add header or footer to dialog.

        Args:
            hf_type: Either header or footer type
            headfoot_lines: Section header or footer (tuple, list, or string with newlines)
        """
        assert hf_type in (CheckerEntryType.HEADER, CheckerEntryType.FOOTER)
        self.new_section()
        if isinstance(headfoot_lines, str):
            headfoot_lines = headfoot_lines.split("\n")
        for line in headfoot_lines:
            self.add_entry(line, None, None, None, hf_type)
        self.new_section()

    def add_header(self, *header_lines: str) -> None:
        """Add header to dialog.

        Args:
            header_lines: strings to add as header lines (strings may contain newlines)
        """
        self._add_headfoot(CheckerEntryType.HEADER, *header_lines)

    def add_footer(self, *footer_lines: str) -> None:
        """Add footer to dialog.

        Args:
            footer_lines: strings to add as footer lines (strings may contain newlines)
        """
        self._add_headfoot(CheckerEntryType.FOOTER, *footer_lines)

    def add_entry(
        self,
        msg: str,
        text_range: Optional[IndexRange] = None,
        hilite_start: Optional[int] = None,
        hilite_end: Optional[int] = None,
        entry_type: CheckerEntryType = CheckerEntryType.CONTENT,
        error_prefix: str = "",
    ) -> None:
        """Add an entry ready to be displayed in the dialog.

        Also set marks in main text at locations of start & end of point of interest.
        Use this for content; use add_header & add_footer for headers & footers.

        Args:
            msg: Entry to be displayed - newline characters are replaced with "⏎".
            text_range: Optional start & end of point of interest in main text widget.
            hilite_start: Optional column to begin higlighting entry in dialog.
            hilite_end: Optional column to end higlighting entry in dialog.
            entry_type: Defaults to content
            error_prefix: Optional string prefix to indicate error
        """
        line = re.sub("\n", "⏎", msg)
        entry = CheckerEntry(
            line,
            text_range,
            hilite_start,
            hilite_end,
            self.section_count,
            len(self.entries),
            entry_type,
            error_prefix,
        )
        self.entries.append(entry)

        if text_range is not None:
            maintext().set_mark_position(
                self.mark_from_rowcol(text_range.start),
                text_range.start,
                gravity=tk.LEFT,
            )
            maintext().set_mark_position(
                self.mark_from_rowcol(text_range.end), text_range.end, gravity=tk.RIGHT
            )

    def display_entries(self, auto_select_line: bool = True) -> None:
        """Display all the stored entries in the dialog according to
        the sort setting.

        Args:
            auto_select_line: Default True. Set to False if calling routine takes
            responsibility for selecting a line in the dialog.
        """

        Busy.busy()
        sort_key: Callable[[CheckerEntry], tuple]
        if (
            self.get_dialog_pref(PrefKey.CHECKERDIALOG_SORT_TYPE_DICT)
            == CheckerSortType.ALPHABETIC
        ):
            sort_key = self.alpha_key
        else:
            sort_key = self.rowcol_key
        self.entries.sort(key=sort_key)
        # By double-clicking button, user can end up with two versions of tool running
        # and then erroring on this line - trap it here and just return, allowing other
        # version of tool to run.
        try:
            self.text.delete("1.0", tk.END)
        except tk.TclError:
            Busy.unbusy()
            return
        self.count_linked_entries = 0
        self.count_suspects = 0

        # Find longest line & col strings to aid formatting
        maxrow = 0
        maxcol = 0
        for entry in self.entries:
            if self.skip_entry(entry):
                continue
            if entry.text_range is not None:
                maxrow = max(maxrow, entry.text_range.start.row)
                maxcol = max(maxcol, entry.text_range.start.col)
        maxrowlen = len(str(maxrow))
        maxcollen = len(str(maxcol)) + 2  # Always colon & at least 1 space after col

        for entry in self.entries:
            if self.skip_entry(entry):
                continue
            rowcol_str = ""
            if entry.severity >= CheckerEntrySeverity.INFO:
                self.count_linked_entries += 1
            if entry.severity >= CheckerEntrySeverity.ERROR:
                self.count_suspects += 1
            if entry.text_range is not None:
                colstr = f"{entry.text_range.start.col}:"
                rowcol_str = (
                    f"{entry.text_range.start.row:>{maxrowlen}}.{colstr:<{maxcollen}}"
                )
            self.text.insert(
                tk.END, rowcol_str + entry.error_prefix + entry.text + "\n"
            )
            if entry.hilite_start is not None and entry.hilite_end is not None:
                start_rowcol = IndexRowCol(self.text.index(tk.END + "-2line"))
                start_rowcol.col = (
                    entry.hilite_start + len(rowcol_str) + len(entry.error_prefix)
                )
                end_rowcol = IndexRowCol(
                    start_rowcol.row,
                    entry.hilite_end + len(rowcol_str) + len(entry.error_prefix),
                )
                self.text.tag_add(
                    HILITE_TAG_NAME, start_rowcol.index(), end_rowcol.index()
                )
            if entry.error_prefix:
                start_rowcol = IndexRowCol(self.text.index(tk.END + "-2line"))
                start_rowcol.col = len(rowcol_str)
                end_rowcol = IndexRowCol(
                    start_rowcol.row, len(rowcol_str) + len(entry.error_prefix)
                )
                self.text.tag_add(
                    ERROR_PREFIX_TAG_NAME, start_rowcol.index(), end_rowcol.index()
                )

        # The default automatic highlighting of previously selected line, or if none
        # the first suitable line, is undesirable when an application immediately
        # selects a different line after displaying the entries. Highlighting will
        # 'jump' before it settles on the newly selected line.
        #
        # To avoid this, an application can disable the default automatic highlighting
        # and handle the selection and highlighting itself; e.g. illo_sn_fixup.py.
        if auto_select_line:
            # Highlight previously selected line, or if none, the first suitable line
            selection_made = False
            if self.selected_text:
                for index, entry in enumerate(self.entries):
                    if (
                        entry.text == self.selected_text
                        and entry.text_range == self.selected_text_range
                        and not self.skip_entry(entry)
                    ):
                        self.select_entry_by_index(index)
                        selection_made = True
                        break
            if not selection_made:
                for index, entry in enumerate(self.entries):
                    if self.skip_entry(entry):
                        continue
                    if entry.text_range:
                        self.select_entry_by_index(index)
                        break
        self.update_count_label()
        Busy.unbusy()

    def showing_suspects_only(self) -> bool:
        """Return whether dialog is showing Suspects Only.

        Returns: True if there's a Suspects Only button that is switched on.
        """
        return self.suspects_only_btn is not None and self.get_dialog_pref(
            PrefKey.CHECKERDIALOG_SUSPECTS_ONLY_DICT
        )

    def skip_entry(self, entry: CheckerEntry) -> bool:
        """Return whether to skip an entry either due to Suspects Only,
        or due to View Options.

        Args:
            entry: Entry to be checked.

        Returns: True if showing Suspects Only but the entry isn't a suspect,
                 or if entry is hidden by View Options settings.
        """
        return (
            self.showing_suspects_only()
            and entry.severity < CheckerEntrySeverity.ERROR
            or self.hide_entry(entry)
        )

    def hide_entry(self, entry: CheckerEntry) -> bool:
        """Return whether to hide entry based upon View Options.

        Args:
            entry: Entry to be checked.

        Returns: True if entry is hidden by View Options settings.
        """
        for view_option_filter in self.view_options_filters:
            if view_option_filter.on and view_option_filter.matches(entry):
                return True
        return False

    def update_count_label(self, working: bool = False) -> None:
        """Update the label showing how many linked entries & suspects are in dialog.

        Args:
            working: If set to True, display a "Working..." label instead.
        """
        if self.count_label.winfo_exists():
            if working:
                self.count_label["text"] = "Working..."
                self.count_label.update()
                return
            es = sing_plur(self.count_linked_entries, "Entry", "Entries")
            ss = (
                f"({sing_plur(self.count_suspects, 'Suspect')})"
                if self.count_suspects > 0
                else ""
            )
            self.count_label["text"] = f"{es} {ss}"

    def select_entry_by_arrow(self, increment: int) -> None:
        """Select next/previous line in dialog, and jump to the line in the
        main text widget that corresponds to it.

        Args:
            increment: +1 to move to next line, -1 to move to previous line.
        """
        line_num = self.text.get_select_line_num()
        if line_num is None:
            return
        try:
            entry_index = self.entry_index_from_linenum(line_num + increment)
        except IndexError:
            return
        self.select_entry_by_index(entry_index, focus=False)

    def select_entry_by_click(self, event: tk.Event) -> str:
        """Select clicked line in dialog, and jump to the line in the
        main text widget that corresponds to it.

        Args:
            event: Event object containing mouse click position.

        Returns:
            "break" to avoid calling other callbacks.
        """
        try:
            entry_index = self.entry_index_from_click(event)
        except IndexError:
            return "break"
        self.select_entry_by_index(entry_index)
        return "break"

    def process_entry_by_click(
        self, event: tk.Event, all_matching: bool = False
    ) -> str:
        """Select clicked line in dialog, and jump to the line in the
        main text widget that corresponds to it. Finally call the
        "process" callback function, if any.

        Args:
            event: Event object containing mouse click position.

        Returns:
            "break" to avoid calling other callbacks.
        """
        try:
            entry_index = self.entry_index_from_click(event)
        except IndexError:
            return "break"
        self.select_entry_by_index(entry_index)
        self.process_entry_current(all_matching=all_matching)
        return "break"

    def remove_entry_by_click(self, event: tk.Event, all_matching: bool = False) -> str:
        """Remove the entry that was clicked in the dialog.

        Args:
            event: Event object containing mouse click position.
            all_matching: If True remove all other entries that have the same
                message as the chosen entry (e.g. same spelling error)

        Returns:
            "break" to avoid calling other callbacks.
        """
        try:
            entry_index = self.entry_index_from_click(event)
        except IndexError:
            return "break"
        self.select_entry_by_index(entry_index)
        self.remove_entry_current(all_matching=all_matching)
        return "break"

    def process_remove_entry_by_click(
        self, event: tk.Event, all_matching: bool = False
    ) -> str:
        """Select clicked line in dialog, and jump to the line in the
        main text widget that corresponds to it. Call the
        "process" callback function, if any, then remove the entry.

        Args:
            event: Event object containing mouse click position.
            all_matching: If True remove all other entries that have the same
                message as the chosen entry (e.g. same spelling error)

        Returns:
            "break" to avoid calling other callbacks.
        """
        try:
            entry_index = self.entry_index_from_click(event)
        except IndexError:
            return "break"
        self.select_entry_by_index(entry_index)
        self.process_remove_entry_current(all_matching=all_matching)
        return "break"

    def entry_index_from_click(self, event: tk.Event) -> int:
        """Get the index into the list of entries based on the mouse position
        in the click event.

        Args:
            event: Event object containing mouse click position.

        Returns:
            Index into self.entries list
            Raises IndexError exception if out of range
        """
        linenum = self.linenum_from_click(event)

        # Convenient place to intercept mouse click and see if user
        # was actually requesting a refresh
        if linenum == 1 and self.text.get("1.0", "1.end") == REFRESH_MESSAGE:
            self.rerun_command()
            self.selection_on_clear[self.__class__.__name__] = None
            raise IndexError  # No valid index selected

        return self.entry_index_from_linenum(linenum)

    def linenum_from_click(self, event: tk.Event) -> int:
        """Get the line number based on the mouse position in the click event.

        Args:
            event: Event object containing mouse click position.

        Returns:
            Line number that was clicked in the list of entries
        """
        return IndexRowCol(self.text.index(f"@{event.x},{event.y}")).row

    def entry_index_from_linenum(self, linenum: int) -> int:
        """Get the index into the entries array from a line number in the list,
        taking into account "Suspects Only" setting.

        Args:
            linenum: Line number in the list of entries.

        Returns:
            Index into entries array.
            Raises IndexError exception if out of range.
        """
        count = 0
        for index, entry in enumerate(self.entries):
            if self.skip_entry(entry):
                continue
            count += 1
            if linenum == count:
                return index
        raise IndexError

    def linenum_from_entry_index(self, entry_index: int) -> int:
        """Get the line number in the list, from the entry_index,
        taking into account "Suspects Only" setting.

        Args:
            entry_index: Index into entries array.

        Returns:
            Line number in the list of entries, or 0 if not in range.
        """
        linenum = 0
        for index, entry in enumerate(self.entries):
            if self.skip_entry(entry):
                continue
            linenum += 1
            if index == entry_index:
                return linenum
        return 0

    def process_remove_entries(
        self, process: bool, remove: bool, all_matching: bool
    ) -> None:
        """Process and/or remove the current entry, if any.

        Args:
            process: If True, process the current entry.
            remove: If True, remove the current entry.
            all_matching: If True, repeat for all entries that have the same
                message as the current entry (e.g. same spelling error).
        """
        entry_index = self.current_entry_index()
        if entry_index is None:
            return
        linenum = self.text.get_select_line_num()
        if linenum is None:
            return
        # Mark before starting so location can be selected later
        self.text.mark_set(MARK_ENTRY_TO_SELECT, f"{linenum}.0")
        match_text = self.entries[entry_index].text
        if all_matching:
            # Work in reverse since may be deleting from list while iterating
            indices = range(len(self.entries) - 1, -1, -1)
        else:
            indices = range(entry_index, entry_index + 1)
        for ii in indices:
            if self.entries[ii].text == match_text:
                if process and self.process_command:
                    self.process_command(self.entries[ii])
                if remove:
                    if self.entries[ii].severity >= CheckerEntrySeverity.INFO:
                        self.count_linked_entries -= 1
                    if self.entries[ii].severity >= CheckerEntrySeverity.ERROR:
                        self.count_suspects -= 1
                    linenum = self.linenum_from_entry_index(ii)
                    self.text.delete(f"{linenum}.0", f"{linenum + 1}.0")
                    del self.entries[ii]
        self.update_count_label()
        # Select line that is now where the first processed/removed line was
        entry_rowcol = IndexRowCol(self.text.index(MARK_ENTRY_TO_SELECT))
        last_row = IndexRowCol(self.text.index(tk.END)).row - 1
        if last_row > 0:
            try:
                entry_index = self.entry_index_from_linenum(
                    min(entry_rowcol.row, last_row)
                )
            except IndexError:
                return
            self.select_entry_by_index(entry_index)

    def process_entry_current(self, all_matching: bool = False) -> None:
        """Call the "process" callback function, if any, on the
        currently selected entry, if any.

        Args:
            all_matching: If True process all other entries that have the same
                message as the chosen entry
        """
        self.process_remove_entries(
            process=True, remove=False, all_matching=all_matching
        )

    def remove_entry_current(self, all_matching: bool = False) -> None:
        """Remove the current entry, if any.

        Args:
            all_matching: If True remove all other entries that have the same
                message as the chosen entry
        """
        self.process_remove_entries(
            process=False, remove=True, all_matching=all_matching
        )

    def process_remove_entry_current(self, all_matching: bool = False) -> None:
        """Call the "process" callback function, if any, for the current entry, if any,
        then remove the entry.

        Args:
            all_matching: If True process & remove all other entries that have the same
                message as the chosen entry
        """
        self.process_remove_entries(
            process=True, remove=True, all_matching=all_matching
        )

    def current_entry_index(self) -> Optional[int]:
        """Get the index entry of the currently selected error message.

        Returns:
            Index into self.entries array, or None if no message selected.
        """
        line_num = self.text.get_select_line_num()
        return None if line_num is None else self.entry_index_from_linenum(line_num)

    def select_entry_by_index(self, entry_index: int, focus: bool = True) -> None:
        """Select line in dialog corresponding to given entry index,
        and jump to the line in the main text widget that corresponds to it.

        Args:
            event: Event object containing mouse click position.
        """
        linenum = self.linenum_from_entry_index(entry_index)
        self.text.select_line(linenum)
        self.text.mark_set(tk.INSERT, f"{linenum}.0")
        self.text.focus_set()
        self.text.tag_remove("sel", "1.0", tk.END)
        try:
            entry = self.entries[entry_index]
        except IndexError:
            self.lift()
            return  # OK if index is no longer valid
        self.selected_text = entry.text
        self.selected_text_range = entry.text_range
        maintext().remove_spotlights()
        if entry.text_range is not None:
            if root().state() == "iconic":
                root().deiconify()
            start = maintext().index(self.mark_from_rowcol(entry.text_range.start))
            end = maintext().index(self.mark_from_rowcol(entry.text_range.end))
            maintext().spotlight_range(IndexRange(start, end))
            maintext().set_insert_index(
                IndexRowCol(start), focus=(focus and self.switch_focus_when_clicked)
            )
            maintext().clear_selection()
        self.lift()
        if (
            self.view_options_dialog is not None
            and self.view_options_dialog.winfo_exists()
        ):
            self.view_options_dialog.lift()

    @classmethod
    def mark_from_rowcol(cls, rowcol: IndexRowCol) -> str:
        """Return name to use to mark given location in text file.

        Args:
            rowcol: Location in text file to be marked.

        Returns:
            Name for mark, e.g. "Checker123.45"
        """
        return f"{cls.get_mark_prefix()}{rowcol.index()}"
